# sage_setup: distribution = sagemath-objects
from sage.structure.sage_object cimport SageObject

cdef class Functor(SageObject):
    cdef __weakref__
    cdef object __domain
    cdef object __codomain



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs scipy sage.symbolic
"""
(Ring-)LWE oracle generators

The Learning with Errors problem (LWE) is solving linear systems of equations
where the right hand side has been disturbed 'slightly' where 'slightly' is made
precise by a noise distribution - typically a discrete Gaussian
distribution. See [Reg09]_ for details.

The Ring Learning with Errors problem (LWE) is solving a set of univariate
polynomial equations - typically in a cyclotomic field - where the right hand
side was disturbed 'slightly'. See [LPR2010]_ for details.

This module implements generators of LWE samples where parameters are chosen
following proposals in the cryptographic literature.

EXAMPLES:

We get 30 samples from an LWE oracle parameterised by security parameter
``n=20`` and where the modulus and the standard deviation of the noise are
chosen as in [Reg09]_::

    sage: from sage.crypto.lwe import samples
    sage: S = samples(30, 20, 'Regev')
    sage: len(S)
    30
    sage: S[0][0].parent(), S[0][1].parent()
    (Vector space of dimension 20 over Ring of integers modulo 401,
     Ring of integers modulo 401)

We may also pass classes to the samples function, which is useful for users
implementing their own oracles::

    sage: from sage.crypto.lwe import samples, LindnerPeikert
    sage: S = samples(30, 20, LindnerPeikert)
    sage: len(S)
    30
    sage: S[0][0].parent(), S[0][1].parent()
    (Vector space of dimension 20 over Ring of integers modulo 2053,
     Ring of integers modulo 2053)

Finally, :func:`samples` also accepts instances of classes::

    sage: from sage.crypto.lwe import LindnerPeikert
    sage: lwe = LindnerPeikert(20)
    sage: S = samples(30, 20, lwe)
    sage: len(S)
    30
    sage: S[0][0].parent(), S[0][1].parent()
    (Vector space of dimension 20 over Ring of integers modulo 2053,
     Ring of integers modulo 2053)

Note that Ring-LWE samples are returned as vectors::

    sage: # needs sage.libs.pari
    sage: from sage.crypto.lwe import RingLWE
    sage: from sage.stats.distributions.discrete_gaussian_polynomial import DiscreteGaussianDistributionPolynomialSampler
    sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], euler_phi(16), 5)
    sage: ringlwe = RingLWE(16, 257, D, secret_dist='uniform')
    sage: p = samples(30, euler_phi(16), ringlwe)[0][0].parent(); p
    Vector space of dimension 8 over Ring of integers modulo 257
    sage: assert all(c.parent() is p for b in samples(30, euler_phi(16), ringlwe) for c in b)

One technical issue when working with these generators is that by default they
return vectors and scalars over/in rings modulo some `q`. These are represented
as elements in `(0,q-1)` by Sage. However, it usually is more natural to think
of these entries as integers in `(-q//2,q//2)`. To allow for this, this module
provides the option to balance the representation. In this case vectors and
scalars over/in the integers are returned::

    sage: from sage.crypto.lwe import samples
    sage: for s in samples(30, 20, 'Regev', balanced=True):
    ....:     s1 = list(s[0]) + [s[1]]
    ....:     assert all(-401//2 <= b <= 401//2 for b in s1)

AUTHORS:

- Martin Albrecht
- Robert Fitzpatrick
- Daniel Cabracas
- Florian Göpfert
- Michael Schneider

REFERENCES:

- [Reg09]_

- [LP2011]_

- [LPR2010]_

- [CGW2013]_
"""

from sage.arith.misc import euler_phi, next_prime
from sage.functions.log import log
from sage.functions.other import floor, ceil
from sage.misc.functional import cyclotomic_polynomial, round
from sage.misc.functional import sqrt
from sage.misc.prandom import randint
from sage.misc.randstate import set_random_seed
from sage.modules.free_module import FreeModule
from sage.modules.free_module_element import random_vector, vector
from sage.numerical.optimize import find_root
from sage.rings.finite_rings.integer_mod_ring import IntegerModRing
from sage.rings.integer_ring import ZZ
from sage.rings.real_mpfr import RR
from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
from sage.stats.distributions.discrete_gaussian_polynomial import DiscreteGaussianDistributionPolynomialSampler
from sage.structure.element import parent
from sage.structure.sage_object import SageObject
from sage.symbolic.constants import pi
from sage.symbolic.ring import SR


class UniformSampler(SageObject):
    """
    Uniform sampling in a range of integers.

    EXAMPLES::

        sage: from sage.crypto.lwe import UniformSampler
        sage: sampler = UniformSampler(-2, 2); sampler
        UniformSampler(-2, 2)
        sage: sampler() in range(-2, 3)
        True

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, lower_bound, upper_bound):
        """
        Construct a uniform sampler with bounds ``lower_bound`` and
        ``upper_bound`` (both endpoints inclusive).

        INPUT:

        - ``lower_bound`` -- integer
        - ``upper_bound`` -- integer

        EXAMPLES::

            sage: from sage.crypto.lwe import UniformSampler
            sage: UniformSampler(-2, 2)
            UniformSampler(-2, 2)
        """
        if lower_bound > upper_bound:
            raise TypeError("lower bound must be <= upper bound.")
        self.lower_bound = ZZ(lower_bound)
        self.upper_bound = ZZ(upper_bound)

    def __call__(self):
        """
        Return a new sample.

        EXAMPLES::

            sage: from sage.crypto.lwe import UniformSampler
            sage: sampler = UniformSampler(-12, 12)
            sage: sampler() in range(-12, 13)
            True
        """
        return randint(self.lower_bound, self.upper_bound)

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import UniformSampler
            sage: UniformSampler(-2, 2)
            UniformSampler(-2, 2)
        """
        return "UniformSampler(%d, %d)" % (self.lower_bound, self.upper_bound)


class UniformPolynomialSampler(SageObject):
    """
    Uniform sampler for polynomials.

    EXAMPLES::

        sage: from sage.crypto.lwe import UniformPolynomialSampler
        sage: UniformPolynomialSampler(ZZ['x'], 8, -2, 2)().parent()
        Univariate Polynomial Ring in x over Integer Ring

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, P, n, lower_bound, upper_bound):
        """
        Construct a sampler for univariate polynomials of degree ``n-1`` where
        coefficients are drawn uniformly at random between ``lower_bound`` and
        ``upper_bound`` (both endpoints inclusive).

        INPUT:

        - ``P`` -- a univariate polynomial ring over the Integers
        - ``n`` -- number of coefficients to be sampled
        - ``lower_bound`` -- integer
        - ``upper_bound`` -- integer

        EXAMPLES::

            sage: from sage.crypto.lwe import UniformPolynomialSampler
            sage: UniformPolynomialSampler(ZZ['x'], 10, -10, 10)
            UniformPolynomialSampler(10, -10, 10)
        """
        self.n = ZZ(n)
        self.P = P
        if lower_bound > upper_bound:
            raise TypeError("lower bound must be <= upper bound.")
        self.lower_bound = ZZ(lower_bound)
        self.upper_bound = ZZ(upper_bound)
        self.D = UniformSampler(self.lower_bound, self.upper_bound)

    def __call__(self):
        """
        Return a new sample.

        EXAMPLES::

            sage: from sage.crypto.lwe import UniformPolynomialSampler
            sage: sampler = UniformPolynomialSampler(ZZ['x'], 8, -12, 12)
            sage: sampler().parent()
            Univariate Polynomial Ring in x over Integer Ring
        """
        coeff = [self.D() for _ in range(self.n)]
        f = self.P(coeff)
        return f

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import UniformPolynomialSampler
            sage: UniformPolynomialSampler(ZZ['x'], 8, -3, 3)
            UniformPolynomialSampler(8, -3, 3)
        """
        return "UniformPolynomialSampler(%d, %d, %d)" % (self.n, self.lower_bound, self.upper_bound)


class LWE(SageObject):
    """
    Learning with Errors (LWE) oracle.

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, n, q, D, secret_dist='uniform', m=None):
        r"""
        Construct an LWE oracle in dimension ``n`` over a ring of order
        ``q`` with noise distribution ``D``.

        INPUT:

        - ``n`` -- dimension (integer > 0)
        - ``q`` -- modulus typically > n (integer > 0)
        - ``D`` -- an error distribution such as an instance of
          :class:`DiscreteGaussianDistributionIntegerSampler` or :class:`UniformSampler`
        - ``secret_dist`` -- distribution of the secret (default: ``'uniform'``); one of

          - ``'uniform'`` -- secret follows the uniform distribution in `\Zmod{q}`
          - ``'noise'`` -- secret follows the noise distribution
          - ``(lb, ub)`` -- the secret is chosen uniformly from ``[lb,...,ub]``
            including both endpoints

        - ``m`` -- number of allowed samples or ``None`` if no such limit exists
          (default: ``None``)

        EXAMPLES:

        First, we construct a noise distribution with standard deviation 3.0::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(3.0)

        Next, we construct our oracle::

            sage: from sage.crypto.lwe import LWE
            sage: lwe = LWE(n=20, q=next_prime(400), D=D); lwe
            LWE(20, 401, Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000, 'uniform', None)

        and sample 1000 samples::

            sage: L = []
            sage: def add_samples():
            ....:     global L
            ....:     L += [lwe() for _ in range(100)]
            sage: add_samples()

        To test the oracle, we use the internal secret to evaluate the samples
        in the secret::

            sage: S = lambda : [ZZ(a.dot_product(lwe._LWE__s) - c) for (a,c) in L]

        However, while Sage represents finite field elements between 0 and q-1
        we rely on a balanced representation of those elements here. Hence, we
        fix the representation and recover the correct standard deviation of the
        noise::

            sage: from numpy import std                                                             # needs numpy
            sage: while abs(std([e if e <= 200 else e-401 for e in S()]) - 3.0) > 0.01:             # needs numpy
            ....:     L = []  # reset L to avoid quadratic behaviour
            ....:     add_samples()

        If ``m`` is not ``None`` the number of available samples is restricted::

            sage: from sage.crypto.lwe import LWE
            sage: lwe = LWE(n=20, q=next_prime(400), D=D, m=30)
            sage: _ = [lwe() for _ in range(30)]
            sage: lwe()  # 31
            Traceback (most recent call last):
            ...
            IndexError: Number of available samples exhausted.
        """
        self.n = ZZ(n)
        self.m = m
        self.__i = 0
        self.K = IntegerModRing(q)
        self.FM = FreeModule(self.K, n)
        self.D = D

        self.secret_dist = secret_dist
        if secret_dist == 'uniform':
            self.__s = random_vector(self.K, self.n)
        elif secret_dist == 'noise':
            self.__s = vector(self.K, self.n, [self.D() for _ in range(n)])
        else:
            try:
                lb, ub = map(ZZ, secret_dist)
                self.__s = vector(self.K, self.n, [randint(lb,ub) for _ in range(n)])
            except (IndexError, TypeError):
                raise TypeError("Parameter secret_dist=%s not understood." % (secret_dist))

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: from sage.crypto.lwe import LWE
            sage: D = DiscreteGaussianDistributionIntegerSampler(3.0)
            sage: lwe = LWE(n=20, q=next_prime(400), D=D); lwe
            LWE(20, 401, Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000, 'uniform', None)

            sage: lwe = LWE(n=20, q=next_prime(400), D=D, secret_dist=(-3, 3)); lwe
            LWE(20, 401, Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000, (-3, 3), None)
        """
        if isinstance(self.secret_dist, str):
            return "LWE(%d, %d, %s, '%s', %s)" % (self.n,self.K.order(),self.D,self.secret_dist, self.m)
        else:
            return "LWE(%d, %d, %s, %s, %s)" % (self.n,self.K.order(),self.D,self.secret_dist, self.m)

    def __call__(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import DiscreteGaussianDistributionIntegerSampler, LWE
            sage: LWE(10, 401, DiscreteGaussianDistributionIntegerSampler(3))()[0].parent()
            Vector space of dimension 10 over Ring of integers modulo 401
            sage: LWE(10, 401, DiscreteGaussianDistributionIntegerSampler(3))()[1].parent()
            Ring of integers modulo 401
        """
        if self.m is not None:
            if self.__i >= self.m:
                raise IndexError("Number of available samples exhausted.")
        self.__i += 1
        a = self.FM.random_element()
        return a, a.dot_product(self.__s) + self.K(self.D())


class Regev(LWE):
    """
    LWE oracle with parameters as in [Reg09]_.

    .. automethod:: __init__
    """
    def __init__(self, n, secret_dist='uniform', m=None):
        """
        Construct LWE instance parameterised by security parameter ``n`` where
        the modulus ``q`` and the ``stddev`` of the noise are chosen as in
        [Reg09]_.

        INPUT:

        - ``n`` -- security parameter (integer > 0)
        - ``secret_dist`` -- distribution of the secret. See documentation of :class:`LWE`
          for details (default='uniform')
        - ``m`` -- number of allowed samples or ``None`` if no such limit exists
          (default: ``None``)

        EXAMPLES::

            sage: from sage.crypto.lwe import Regev
            sage: Regev(n=20)
            LWE(20, 401, Discrete Gaussian sampler over the Integers with sigma = 1.915069 and c = 401.000000, 'uniform', None)
        """
        q = ZZ(next_prime(n**2))
        s = RR(1/(RR(n).sqrt() * log(n, 2)**2) * q)
        D = DiscreteGaussianDistributionIntegerSampler(s/sqrt(2*pi.n()), q)
        LWE.__init__(self, n=n, q=q, D=D, secret_dist=secret_dist, m=m)


class LindnerPeikert(LWE):
    """
    LWE oracle with parameters as in [LP2011]_.

    .. automethod:: __init__
    """
    def __init__(self, n, delta=0.01, m=None):
        """
        Construct LWE instance parameterised by security parameter ``n`` where
        the modulus ``q`` and the ``stddev`` of the noise is chosen as in
        [LP2011]_.

        INPUT:

        - ``n`` -- security parameter (integer > 0)
        - ``delta`` -- error probability per symbol (default: 0.01)
        - ``m`` -- number of allowed samples or ``None`` in which case ``m=2*n +
          128`` as in [LP2011]_ (default: ``None``)

        EXAMPLES::

            sage: from sage.crypto.lwe import LindnerPeikert
            sage: LindnerPeikert(n=20)
            LWE(20, 2053, Discrete Gaussian sampler over the Integers with sigma = 3.600954 and c = 0.000000, 'noise', 168)
        """
        if m is None:
            m = 2*n + 128
        # Find c>=1 such that c*exp((1-c**2)/2))**(2*n) == 2**-40
        #         (c*exp((1-c**2)/2))**(2*n) == 2**-40
        #    log((c*exp((1-c**2)/2))**(2*n)) == -40*log(2)
        #       (2*n)*log(c*exp((1-c**2)/2)) == -40*log(2)
        #  2*n*(log(c)+log(exp((1-c**2)/2))) == -40*log(2)
        #            2*n*(log(c)+(1-c**2)/2) == -40*log(2)
        #              2*n*log(c)+n*(1-c**2) == -40*log(2)
        #  2*n*log(c)+n*(1-c**2) + 40*log(2) == 0
        c = SR.var('c')
        c = find_root(2*n*log(c)+n*(1-c**2) + 40*log(2) == 0, 1, 10)
        # Upper bound on s**2/t
        s_t_bound = (sqrt(2) * pi / c / sqrt(2*n*log(2/delta))).n()
        # Interpretation of "choose q just large enough to allow for a Gaussian parameter s>=8" in [LP2011]_
        q = next_prime(floor(2**round(log(256 / s_t_bound, 2))))
        # Gaussian parameter as defined in [LP2011]_
        s = sqrt(s_t_bound*floor(q/4))
        # Transform s into stddev
        stddev = s/sqrt(2*pi.n())
        D = DiscreteGaussianDistributionIntegerSampler(stddev)
        LWE.__init__(self, n=n, q=q, D=D, secret_dist='noise', m=m)


class UniformNoiseLWE(LWE):
    """
    LWE oracle with uniform secret with parameters as in [CGW2013]_.

    .. automethod:: __init__
    """
    def __init__(self, n, instance='key', m=None):
        """
        Construct LWE instance parameterised by security parameter ``n`` where
        all other parameters are chosen as in [CGW2013]_.

        INPUT:

        - ``n`` -- security parameter (integer >= 89)
        - ``instance`` -- one of

          - ``'key'`` -- the LWE-instance that hides the secret key is generated
          - ``'encrypt'`` -- the LWE-instance that hides the message is generated
            (default: ``'key'``)

        - ``m`` -- number of allowed samples or ``None`` in which case ``m`` is
          chosen as in [CGW2013]_.  (default: ``None``)

        EXAMPLES::

            sage: from sage.crypto.lwe import UniformNoiseLWE
            sage: UniformNoiseLWE(89)
            LWE(89, 64311834871, UniformSampler(0, 6577), 'noise', 131)

            sage: UniformNoiseLWE(89, instance='encrypt')
            LWE(131, 64311834871, UniformSampler(0, 11109), 'noise', 181)
        """

        if n < 89:
            raise TypeError("Parameter too small")

        n2 = n
        C = 4/sqrt(2*pi)
        kk = floor((n2-2*log(n2, 2)**2)/5)
        n1 = (3*n2-5*kk) // 2
        ke = floor((n1-2*log(n1, 2)**2)/5)
        l = (3*n1-5*ke) // 2 - n2
        sk = ceil((C*(n1+n2))**(ZZ(3)/2))
        se = ceil((C*(n1+n2+l))**(ZZ(3)/2))
        q = next_prime(max(ceil((4*sk)**(ZZ(n1+n2)/n1)),
                           ceil((4*se)**(ZZ(n1+n2+l)/(n2+l))),
                           ceil(4*(n1+n2)*se*sk+4*se+1)))

        if kk <= 0:
            raise TypeError("Parameter too small")

        if instance == 'key':
            D = UniformSampler(0, sk-1)
            if m is None:
                m = n1
            LWE.__init__(self, n=n2, q=q, D=D, secret_dist='noise', m=m)
        elif instance == 'encrypt':
            D = UniformSampler(0, se-1)
            if m is None:
                m = n2+l
            LWE.__init__(self, n=n1, q=q, D=D, secret_dist='noise', m=m)
        else:
            raise TypeError("Parameter instance=%s not understood." % (instance))


class RingLWE(SageObject):
    """
    Ring Learning with Errors oracle.

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, N, q, D, poly=None, secret_dist='uniform', m=None):
        """
        Construct a Ring-LWE oracle in dimension ``n=phi(N)`` over a ring of order
        ``q`` with noise distribution ``D``.

        INPUT:

        - ``N`` -- index of cyclotomic polynomial (integer > 0, must be power of 2)
        - ``q`` -- modulus typically > N (integer > 0)
        - ``D`` -- an error distribution such as an instance of
          :class:`DiscreteGaussianDistributionPolynomialSampler` or :class:`UniformSampler`
        - ``poly`` -- a polynomial of degree ``phi(N)``. If ``None`` the
          cyclotomic polynomial used (default: ``None``).
        - ``secret_dist`` -- distribution of the secret. See documentation of
          :class:`LWE` for details (default='uniform')
        - ``m`` -- number of allowed samples or ``None`` if no such limit exists
          (default: ``None``)

        EXAMPLES::

            sage: from sage.crypto.lwe import RingLWE
            sage: from sage.stats.distributions.discrete_gaussian_polynomial import DiscreteGaussianDistributionPolynomialSampler
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], n=euler_phi(20), sigma=3.0)                # needs sage.libs.pari
            sage: RingLWE(N=20, q=next_prime(800), D=D)                                 # needs sage.libs.pari
            RingLWE(20, 809, Discrete Gaussian sampler for polynomials of degree < 8 with σ=3.000000 in each component, x^8 - x^6 + x^4 - x^2 + 1, 'uniform', None)
        """
        self.N = ZZ(N)
        self.n = euler_phi(N)
        self.m = m
        self.__i = 0
        self.K = IntegerModRing(q)

        if self.n != D.n:
            raise ValueError("Noise distribution has dimensions %d != %d" % (D.n, self.n))

        self.D = D
        self.q = q
        if poly is not None:
            self.poly = poly
        else:
            self.poly = cyclotomic_polynomial(self.N, 'x')

        self.R_q = self.K['x'].quotient(self.poly, 'x')

        self.secret_dist = secret_dist
        if secret_dist == 'uniform':
            self.__s = self.R_q.random_element()  # uniform sampling of secret
        elif secret_dist == 'noise':
            self.__s = self.D()
        else:
            raise TypeError("Parameter secret_dist=%s not understood." % (secret_dist))

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import DiscreteGaussianDistributionPolynomialSampler, RingLWE
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], n=8, sigma=3.0)
            sage: RingLWE(N=16, q=next_prime(400), D=D)
            RingLWE(16, 401, Discrete Gaussian sampler for polynomials of degree < 8 with σ=3.000000 in each component, x^8 + 1, 'uniform', None)
        """
        if isinstance(self.secret_dist, str):
            return "RingLWE(%d, %d, %s, %s, '%s', %s)" % (self.N, self.K.order(), self.D, self.poly, self.secret_dist, self.m)
        else:
            return "RingLWE(%d, %d, %s, %s, %s, %s)" % (self.N, self.K.order(), self.D, self.poly, self.secret_dist, self.m)

    def __call__(self):
        """
        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: from sage.crypto.lwe import DiscreteGaussianDistributionPolynomialSampler, RingLWE
            sage: N = 16
            sage: n = euler_phi(N)
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], n, 5)
            sage: ringlwe = RingLWE(N, 257, D, secret_dist='uniform')
            sage: ringlwe()[0].parent()
            Vector space of dimension 8 over Ring of integers modulo 257
            sage: ringlwe()[1].parent()
            Vector space of dimension 8 over Ring of integers modulo 257
        """
        if self.m is not None:
            if self.__i >= self.m:
                raise IndexError("Number of available samples exhausted.")
        self.__i += 1
        a = self.R_q.random_element()
        return vector(a), vector(a * (self.__s) + self.D())


class RingLindnerPeikert(RingLWE):
    """
    Ring-LWE oracle with parameters as in [LP2011]_.

    .. automethod:: __init__
    """
    def __init__(self, N, delta=0.01, m=None):
        """
        Construct a Ring-LWE oracle in dimension ``n=phi(N)`` where
        the modulus ``q`` and the ``stddev`` of the noise is chosen as in
        [LP2011]_.

        INPUT:

        - ``N`` -- index of cyclotomic polynomial (integer > 0, must be power of 2)
        - ``delta`` -- error probability per symbol (default: 0.01)
        - ``m`` -- number of allowed samples or ``None`` in which case ``3*n`` is
          used (default: ``None``)

        EXAMPLES::

            sage: from sage.crypto.lwe import RingLindnerPeikert
            sage: RingLindnerPeikert(N=16)
            RingLWE(16, 1031, Discrete Gaussian sampler for polynomials of degree < 8 with σ=2.803372 in each component, x^8 + 1, 'noise', 24)
        """
        n = euler_phi(N)
        if m is None:
            m = 3*n
        # Find c>=1 such that c*exp((1-c**2)/2))**(2*n) == 2**-40
        #  i.e c>=1 such that 2*n*log(c)+n*(1-c**2) + 40*log(2) == 0
        c = SR.var('c')
        c = find_root(2*n*log(c)+n*(1-c**2) + 40*log(2) == 0, 1, 10)
        # Upper bound on s**2/t
        s_t_bound = (sqrt(2) * pi / c / sqrt(2*n*log(2/delta))).n()
        # Interpretation of "choose q just large enough to allow for a Gaussian parameter s>=8" in [LP2011]_
        q = next_prime(floor(2**round(log(256 / s_t_bound, 2))))
        # Gaussian parameter as defined in [LP2011]_
        s = sqrt(s_t_bound*floor(q/4))
        # Transform s into stddev
        stddev = s/sqrt(2*pi.n())
        D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], n, stddev)
        RingLWE.__init__(self, N=N, q=q, D=D, poly=None, secret_dist='noise', m=m)


class RingLWEConverter(SageObject):
    """
    Wrapper callable to convert Ring-LWE oracles into LWE oracles by
    disregarding the additional structure.

    .. automethod:: __init__
    .. automethod:: __call__
    """
    def __init__(self, ringlwe):
        """
        INPUT:

        - ``ringlwe`` -- an instance of a :class:`RingLWE`

        EXAMPLES::

            sage: from sage.crypto.lwe import DiscreteGaussianDistributionPolynomialSampler, RingLWE, RingLWEConverter
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], euler_phi(16), 5)
            sage: lwe = RingLWEConverter(RingLWE(16, 257, D, secret_dist='uniform'))
            sage: set_random_seed(1337)
            sage: lwe()
            ((171, 197, 58, 125, 3, 216, 32, 130), ...)
        """
        self.ringlwe = ringlwe
        self._i = 0
        self._ac = None
        self.n = self.ringlwe.n

    def __call__(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import DiscreteGaussianDistributionPolynomialSampler, RingLWE, RingLWEConverter
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], euler_phi(16), 5)
            sage: lwe = RingLWEConverter(RingLWE(16, 257, D, secret_dist='uniform'))
            sage: set_random_seed(1337)
            sage: lwe()
            ((171, 197, 58, 125, 3, 216, 32, 130), ...)
        """
        R_q = self.ringlwe.R_q

        if (self._i % self.n) == 0:
            self._ac = self.ringlwe()
        a, c = self._ac
        x = R_q.gen()
        r = vector((x**(self._i % self.n) * R_q(a.list())).list()), c[self._i % self.n]
        self._i += 1
        return r

    def _repr_(self):
        """
        EXAMPLES::

            sage: from sage.crypto.lwe import DiscreteGaussianDistributionPolynomialSampler, RingLWE, RingLWEConverter
            sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], euler_phi(20), 5)
            sage: rlwe = RingLWE(20, 257, D)
            sage: lwe = RingLWEConverter(rlwe)
            sage: lwe
            RingLWEConverter(RingLWE(20, 257, Discrete Gaussian sampler for polynomials of degree < 8 with σ=5.000000 in each component, x^8 - x^6 + x^4 - x^2 + 1, 'uniform', None))
        """
        return "RingLWEConverter(%s)" % str(self.ringlwe)


def samples(m, n, lwe, seed=None, balanced=False, **kwds):
    """
    Return ``m`` LWE samples.

    INPUT:

    - ``m`` -- the number of samples (integer > 0)
    - ``n`` -- the security parameter (integer > 0)
    - ``lwe`` -- either

      - a subclass of :class:`LWE` such as :class:`Regev` or :class:`LindnerPeikert`
      - an instance of :class:`LWE` or any subclass
      - the name of any such class (e.g., "Regev", "LindnerPeikert")

    - ``seed`` -- seed to be used for generation or ``None`` if no specific seed
      shall be set (default: ``None``)
    - ``balanced`` -- use function :func:`balance_sample` to return balanced
      representations of finite field elements (default: ``False``)
    - ``**kwds`` -- passed through to LWE constructor

    EXAMPLES::

        sage: from sage.crypto.lwe import samples, Regev
        sage: samples(2, 20, Regev, seed=1337)
        [((199, 388, 337, 53, 200, 284, 336, 215, 75, 14, 274, 234, 97, 255, 246, 153, 268, 218, 396, 351), 15),
         ((365, 227, 333, 165, 76, 328, 288, 206, 286, 42, 175, 155, 190, 275, 114, 280, 45, 218, 304, 386), 143)]

        sage: from sage.crypto.lwe import samples, Regev
        sage: samples(2, 20, Regev, balanced=True, seed=1337)
        [((199, -13, -64, 53, 200, -117, -65, -186, 75, 14, -127, -167, 97, -146, -155, 153, -133, -183, -5, -50), 15),
         ((-36, -174, -68, 165, 76, -73, -113, -195, -115, 42, 175, 155, 190, -126, 114, -121, 45, -183, -97, -15), 143)]

        sage: from sage.crypto.lwe import samples
        sage: samples(2, 20, 'LindnerPeikert')
        [((506, 1205, 398, 0, 337, 106, 836, 75, 1242, 642, 840, 262, 1823, 1798, 1831, 1658, 1084, 915, 1994, 163), 1447),
         ((463, 250, 1226, 1906, 330, 933, 1014, 1061, 1322, 2035, 1849, 285, 1993, 1975, 864, 1341, 41, 1955, 1818, 1357), 312)]
    """
    if seed is not None:
        set_random_seed(seed)

    if isinstance(lwe, str):
        lwe = eval(lwe)

    if isinstance(lwe, type):
        lwe = lwe(n, m=m, **kwds)
    else:
        if lwe.n != n:
            raise ValueError("Passed LWE instance has n=%d, but n=%d was passed to this function." % (lwe.n, n))

    if balanced is False:
        f = lambda a_c: a_c
    else:
        f = balance_sample
    return [f(lwe()) for _ in range(m)]


def balance_sample(s, q=None):
    r"""
    Given ``(a,c) = s`` return a tuple ``(a',c')`` where ``a'`` is an integer
    vector with entries between -q//2 and q//2 and ``c`` is also within these
    bounds.

    If ``q`` is given ``(a,c) = s`` may live in the integers. If ``q`` is not
    given, then ``(a,c)`` are assumed to live in `\Zmod{q}`.

    INPUT:

    - ``s`` -- sample of the form (a,c) where a is a vector and c is a scalar
    - ``q`` -- modulus (default: ``None``)

    EXAMPLES::

        sage: from sage.crypto.lwe import balance_sample, samples, Regev
        sage: for s in samples(10, 5, Regev):
        ....:     b = balance_sample(s)
        ....:     assert all(-29//2 <= c <= 29//2 for c in b[0])
        ....:     assert -29//2 <= b[1] <= 29//2
        ....:     assert all(s[0][j] == b[0][j] % 29 for j in range(5))
        ....:     assert s[1] == b[1] % 29


        sage: from sage.crypto.lwe import balance_sample, DiscreteGaussianDistributionPolynomialSampler, RingLWE, samples
        sage: D = DiscreteGaussianDistributionPolynomialSampler(ZZ['x'], 8, 5)
        sage: rlwe = RingLWE(20, 257, D)
        sage: for s in samples(10, 8, rlwe):
        ....:     b = balance_sample(s)
        ....:     assert all(-257//2 <= c <= 257//2 for bi in b for c in bi)
        ....:     assert all(s[i][j] == b[i][j] % 257 for i in range(2) for j in range(8))

    .. NOTE::

        This function is useful to convert between Sage's standard
        representation of elements in `\Zmod{q}` as integers between 0 and q-1
        and the usual representation of such elements in lattice cryptography as
        integers between -q//2 and q//2.
    """
    a, c = s

    try:
        c[0]
        scalar = False
    except TypeError:
        c = vector(c.parent(),[c])
        scalar = True

    if q is None:
        q = parent(c[0]).order()
        a = a.change_ring(ZZ)
        c = c.change_ring(ZZ)
    else:
        K = IntegerModRing(q)
        a = a.change_ring(K).change_ring(ZZ)
        c = c.change_ring(K).change_ring(ZZ)

    q2 = q//2

    if scalar:
        return vector(ZZ, len(a), [e if e <= q2 else e-q for e in a]), c[0] if c[0] <= q2 else c[0]-q
    else:
        return vector(ZZ, len(a), [e if e <= q2 else e-q for e in a]), vector(ZZ, len(c), [e if e <= q2 else e-q for e in c])



# sage_setup: distribution = sagemath-modules
r"""
Implementation of the command ``sage --fiximports``.

This file provides a tool to fix the modularization antipattern ``namespace_pkg_all_import``
reported by ``tox -e relint``. Sage library code should not import from ``sage.PAC.KAGE.all``
when ``sage.PAC.KAGE`` is an implicit namespace package.

AUTHORS:

- Alex Chandler
- Matthias Köppe

INSTRUCTIONS:

To fix the above issue for all Python and Cython source files (``.py``, ``.pyx``, ``.pxi``) files in the ``src/sage`` directory,
run the following from a terminal in ``SAGE_ROOT`` ::

    ./sage -python src/sage/misc/replace_dot_all.py

or ::

    ./sage --fiximports

Running replace_dot_all.py will call the function :func:`walkdir_replace_dot_all` which walks through all Python and Cython source files
and replaces certain ``from sage.PAC.KAGE.all import something`` (those matching the pattern above)
with the correct ``import`` statement by applying the function :func:`~sage.misc.dev_tools.import_statements`.

The user can also pass subdirectories of ``src/sage`` or specific files to fix. For example ::

    ./sage -python src/sage/misc/replace_dot_all.py src/sage/arith

will fix all files in ``src/sage/arith`` and ::

    ./sage -python src/sage/misc/replace_dot_all.py src/sage/arith/functions.pyx

will fix the file ``src/sage/arith/functions.py``. The file extension is necessary in the case of a specific file. The user can also
pass the verbose flag ``-v`` to print out the files being fixed. For example ::

    ./sage -python src/sage/misc/replace_dot_all.py -v src/sage/arith

will fix all files in ``src/sage/arith`` and print out the unusual examples of ``import`` statements it finds.

In some rare cases, such as ``import`` statements appearing in doctests, the program will not be able to fix the ``import`` statement. The program will
print out the location of the file and the line number of the exceptional ``import`` statement. The user can then manually fix the ``import`` statement.
The program will also (usually) print out the suggested replacement for the ``import`` statement. The user can then copy and paste this replacement
into the file. In the cases a suggested replacement is not printed out, the user should use the function :func:`~sage.misc.dev_tools.import_statements`
to find the correct ``import`` statement.
"""

# ****************************************************************************
#       Copyright (C) 2022-2023 Alex Chandler
#                     2023      Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

# Importing packages

from sage.misc.dev_tools import import_statements
import os
import re
import argparse

# We import this using __import__ so that "tox -e relint" does not complain about this source file.
__import__("sage.all", globals(), locals(), ["*"])


# Keep in sync with SAGE_ROOT/src/.relint.yml (namespace_pkg_all_import)

default_package_regex = (r"sage("
                         r"|[.](arith|categories|combinat|crypto|databases|data_structures|dynamics|ext|game_theory|games|geometry|graphs|groups|interfaces|lfunctions|manifolds|matrix|matroids|misc|modules|monoids|numerical|probability|quadratic_forms|quivers|rings|sat|schemes|sets|stats|tensor)[a-z0-9_.]*|[.]libs"
                         r")[.]all")


# Global variables

examples = list('ABCDEFGHIJ')  # controls how we print out interesting examples to the console
interesting_examples = dict(zip(examples, [0]*len(examples)))
log_messages = ''
number_examples_to_print = 100  # controls how many examples we print out to the console (100 effectively allows all unusual examples to be printed)
numberFiles, numberFilesMatchingRegex, numberFilesChanged, numberStatementsReplaced = 0, 0, 0, 0  # to print report on number of files changed


# Functions

def find_replacements(location, package_regex=None, verbose=False):
    r"""
    Locate the lines in the file at ``location`` which contain an ``import`` statement.

    INPUT:

    - ``location`` -- a file path
    - ``package_regex`` -- (default: :obj:`default_package_regex`) a regular expression matching
      the ``sage.PAC.KAGE.all`` package names from which we do not want to import.
    - ``verbose`` -- a parameter which if used will issue print statements when interesting examples are found

    OUTPUT:

    an array [row_index,import_index,replaced_commands,lines_spanned (optional)] with entries

    - ``row_index`` -- the row index (zero indexed) in the file which needs replacing
    - ``import_index`` -- the index of row row_index which marks the beginning of the word ``import`` in the line
    - ``replaced_commands`` -- the string which will replace the current line
    - ``lines_spanned`` -- the number of lines the original statement spans

    This output can be processed by the function :func:`process_line`.

    EXAMPLES::

        sage: # needs SAGE_SRC
        sage: from sage.misc.replace_dot_all import *
        sage: location = os.path.join(sage.env.SAGE_SRC, 'sage', 'plot', 'arc.py')
        sage: find_replacements(location, package_regex='sage[.]plot[.]all', verbose=True)
        [[..., ..., 'from sage.plot.graphics import Graphics']]
    """
    if package_regex is None:
        package_regex = default_package_regex
    regex = r"from\s+" + package_regex + r"\s+import"
    pattern = re.compile(regex)
    replacements = []
    global log_messages, interesting_examples
    with open(location) as fp:
        skip_line = False
        lines = fp.readlines()  # read all lines using readline()
        row_index = 0
        for row in lines:  # iterate over all lines of python file
            if pattern.search(row):  # (match the regex also do not want to mess with documentation)
                prefix = ''
                if '*' in row or 'SAGE_ROOT' in row:
                    if verbose and interesting_examples['J'] < number_examples_to_print:
                        interesting_examples['J'] += 1
                        log_messages += (f'J. Match but no changes made (import statement uses *) at {location}:{row_index + 1}. '
                                         f'Not applying any changes here.\n')
                    continue
                elif not (row.lstrip()[0:4] == 'from'):
                    skip_line = True
                    if '"' not in row and "'" not in row:
                        print(f'\n'
                              f'NEED TO CHANGE MANUALLY \n'
                              f'  Issue: line with import statement does not start with "from" \n'
                              f'  Location: at {location} \n'
                              f'  Line number: {row_index + 1}. \n'
                              f'  Giving correct import statements:\n')
                        leading_space = 0
                        while len(row) > 0 and row[leading_space] == ' ' and leading_space < len(row)-1:
                            leading_space += 1
                        prefix_space = leading_space
                        while row[prefix_space:prefix_space+4] != 'from':
                            prefix_space += 1
                        prefix = row[leading_space:prefix_space]
                        row = row[prefix_space:]
                    else:
                        print(f'\n'
                              f'NEED TO CHANGE MANUALLY \n'
                              f'  Issue: import statement does not start with "from" and contains quotation marks \n'
                              f'  Location: at {location}:{row_index + 1}. \n'
                              f'  Not able to suggest correct import statements. User must use the function import_statements().')
                        continue
                # find() method returns -1 if the value is not found, or if found it returns index of the first occurrence of the substring
                import_index = row.find('import ')
                modules = ''
                to_exec = row.strip()  # the import statement itself which we will clean and call to import modules e.g. "import (aa as a, bb, cc as c)"
                to_eval = row[import_index + 7:-1].strip()  # tuple of modules in import statement but e.g. "(aa as a, bb, cc as c)" is saved as "(a,bb,c)"
                to_eval_raw = row[import_index + 7:-1].strip()  # same as to_eval but we don't get rid of " as " parts e.g. "(aa as a, bb, cc as c)"
                span = 0  # keeps track of how many lines the import statement spans

                if '(' in row:  # for examples where we import a tuple of modules and the statement spans several lines
                    while ')' not in lines[row_index + span]:  # finding the line which closes the import statement
                        span += 1
                        to_exec += lines[row_index + span].strip()
                        to_eval += lines[row_index + span].strip()
                        to_eval_raw += lines[row_index + span].strip()
                    if span and verbose:  # useful to see these multiline examples for debugging
                        if " as " in to_eval_raw and interesting_examples['D'] < number_examples_to_print:
                            log_messages += f'D. Interesting example (spans multiple lines and has " as ") at {location}:{row_index + 1}\n'
                            interesting_examples['D'] += 1
                        elif interesting_examples['B'] < number_examples_to_print:
                            log_messages += f'B. Interesting example (spans multiple lines) at {location}:{row_index + 1}\n'
                            interesting_examples['B'] += 1

                # if there is an "as" statement inside to_eval, we want to keep only the new name for the module e.g. "(aa as a, bb, cc as c)" becomes "(a,bb,c)"
                while " as " in to_eval:
                    as_ind = to_eval.find(" as ")
                    j = as_ind - 1
                    while to_eval[j] not in [',', ' '] and j >= 0:
                        j -= 1
                    to_eval = to_eval[0:j+1] + to_eval[as_ind+4:]

                try:  # trying to execute the import statement so we can eval the modules and feed them to the function import_statements
                    to_exec = to_exec.replace("'", '').replace('"', '')
                    if (to_exec[-1] == ','):
                        to_exec = to_exec[:-1]
                    glob = dict()
                    exec(to_exec, glob)
                except ModuleNotFoundError as err:
                    print(f'ModuleNotFoundError: {err} found when trying to execute {to_exec}')
                except ImportError as err:
                    print(f'ImportError: {err} found when trying to execute {to_exec}')

                try:  # try to evaluate the list of module names to get a list of the modules themselves which we can call import_statements on
                    modules = eval(to_eval, glob)
                except NameError as err:
                    print(f'NameError: {err} found when trying to evaluate {to_eval} at {location}:{row_index + 1}')
                except SyntaxError as err:
                    print(f'SyntaxError: {err} found when trying to evaluate {to_eval} at {location}:{row_index + 1}')

                # Need module to be a list of modules we are importing. If a single module was given, we make it a 1-element list.
                if not isinstance(modules, tuple):
                    modules = [modules]

                to_eval_list = to_eval.replace('(', '').replace(')', '').split(',')  # convert comma separated string to_eval to a list
                to_eval_list_raw = to_eval_raw.replace('(', '').replace(')', '').split(',')  # convert comma separated string to_eval_raw to a list
                to_eval_list_index = 0

                change_to = ''
                # constructs the appropriate replacement for the import statement and stores it (along with location data) in list replacements
                for mod in modules:
                    postfix = ''
                    as_index = -1
                    # saves the callable name of module in variable postfix (e.g. for module "b" called by "bb as b" we set postfix = " as b")
                    if " as " in to_eval_list_raw[to_eval_list_index]:
                        if verbose and interesting_examples['C'] < number_examples_to_print:
                            log_messages += f'C. Interesting example (" as " in tuple import) at {location}:{row_index + 1}\n'
                            interesting_examples['C'] += 1
                        as_index = to_eval_list_raw[to_eval_list_index].index(" as ")
                        postfix = to_eval_list_raw[to_eval_list_index][as_index:]
                    new_import_statement = import_statements(mod, answer_as_str=True, verbose=False)  # import statement for the current mod in the list module
                    import_index = new_import_statement.find('import')
                    new_mod_as_string = new_import_statement[import_index + 7:].strip()  # the name for the module given by the function import_statements
                    if as_index >= 0:
                        # the name for the module as originally called in the document (when there is an " as " statement)
                        original_mod_string = to_eval_list_raw[to_eval_list_index].strip()[:as_index]
                    else:
                        original_mod_string = to_eval_list[to_eval_list_index].strip()  # the name for the module as originally called in the document
                    if original_mod_string != new_mod_as_string:  # if the names differ, we use the original name as it was called in the document
                        if verbose and interesting_examples['A'] < number_examples_to_print:
                            log_messages += (f'A. Interesting example (module has multiple names) at {location}:{row_index + 1}. '
                                             f'Names: {original_mod_string}, {new_mod_as_string}. '
                                             f'Replacing new {new_mod_as_string} by original {original_mod_string}.\n')
                            interesting_examples['A'] += 1
                        new_import_statement = new_import_statement.replace(' ' + new_mod_as_string, ' ' + new_mod_as_string + ' as ' + original_mod_string)
                        if " as " in postfix and interesting_examples['G'] < number_examples_to_print:
                            log_messages += (f'G. Interesting example (module has multiple names) at {location}:{row_index + 1}. '
                                             f'Names: {original_mod_string}, {new_mod_as_string}. '
                                             f'Replacing new {new_mod_as_string} by original {original_mod_string}.\n')
                            interesting_examples['G'] += 1
                    if len(postfix.strip()) > 0:  # if module was called with " as " statement, we put that back in by adding the string "postfix"
                        # if " as " in new_import_statement locate the index of " as ", remove the end after this, and add the postfix there
                        if " as " in new_import_statement:
                            new_import_statement = new_import_statement[:new_import_statement.index(" as ")] + ' ' + postfix.strip()
                        else:
                            new_import_statement += (' ' + postfix.strip())
                    change_to += (prefix + new_import_statement + '\n')
                    to_eval_list_index += 1
                # [:-1] on change_to gets rid of the last '\n' we added which adds an unnecessary new line
                replacement = [row_index, import_index, change_to[:-1]].copy()
                if span:
                    # if original statement spanned multiple lines, we store that information to signal that we need to skip lines
                    # as we read the document in the function make_replacements_in_file
                    replacement.append(span)
                if not skip_line:
                    replacements.append(replacement)
            row_index += 1
            skip_line = False
    # keeping track of the numbers of files changed and statements replaced
    global numberStatementsReplaced, numberFilesChanged
    numberStatementsReplaced += len(replacements)
    if replacements:
        numberFilesChanged += 1
    return replacements


def process_line(location, line, replacements, row_index, verbose=False):
    r"""
    Modify a single source code ``line`` according to the given ``replacements``.

    INPUTS:

    - ``location`` -- a file path; only used for logging
    - ``line`` -- a source code line
    - ``replacements`` -- the array output from :func:`find_replacements`
    - ``row_index`` -- the line number where ``import`` appears
    - ``verbose`` -- if ``True``, issue print statements when interesting
      examples are found

    OUTPUT: an array ``[new_line, replacements]`` with entries

    - ``new_line`` -- the modified import statement (possibly now on several lines)
    - ``replacements`` -- just returns the original replacements with its index 0 element removed if ``replacements`` is nonempty

    EXAMPLES:

    Replacing the first line which needs a replacement in the source file with filepath ``src/sage/plot/arc.py``::

        sage: # needs SAGE_SRC
        sage: from sage.misc.replace_dot_all import *
        sage: location = os.path.join(sage.env.SAGE_SRC, 'sage', 'plot', 'arc.py')
        sage: replacements = find_replacements(location, package_regex='sage[.]plot[.]all', verbose=True); replacements
        [[477, 24, 'from sage.plot.graphics import Graphics']]
        sage: with open(location, "r") as file:
        ....:     lines = file.readlines()
        sage: row_index, col_number, *_ = replacements[0]
        sage: line = lines[row_index]
        sage: print(line.rstrip())
        from sage.plot.all import Graphics
        sage: new_line, replacements = process_line(location, line, replacements, row_index)
        sage: print(new_line)
        from sage.plot.graphics import Graphics
        sage: replacements
        []
    """
    line = line.rstrip()  # stripping line break
    new_line = ''
    global log_messages, interesting_examples
    if len(replacements) == 0:
        return line, replacements
    if row_index == replacements[0][0]:  # if line marked as containing .all
        replacement = replacements.pop(0)
        leading_space = 0
        while line and line[leading_space] == ' ' and leading_space < len(line)-1:
            leading_space += 1
        new_line = ' '*leading_space + replacement[2]  # adds leading space to first line (which may or may not start with 'from')
        new_line = new_line.replace('\n', '\n'+' '*leading_space)  # adds correct amount of indentation to the replacement at each line
        # new_line = replacement[2].replace('from ',' '*leading_space + 'from ') # adds correct amount of indentation to the replacement at each line
        if verbose and leading_space > 0:
            if len(replacement) == 4 and interesting_examples['F'] < number_examples_to_print:
                log_messages += f'F. Interesting example (has leading space and multiline) at {location}:{replacement[0] + 1}\n'
                interesting_examples['F'] += 1
            elif interesting_examples['E'] < number_examples_to_print:
                log_messages += f'E. Interesting example (has leading space) at {location}:{replacement[0] + 1}\n'
                interesting_examples['E'] += 1

    else:  # if line does not contain .all
        new_line = line
    return new_line, replacements


def make_replacements_in_file(location, package_regex=None, verbose=False, output=None):
    r"""
    Replace ``import`` statements in the file with filepath "location".

    INPUT:

    - ``location`` -- a file path
    - ``package_regex`` -- (default: :obj:`default_package_regex`) a regular expression matching
      the ``sage.PAC.KAGE.all`` package names from which we do not want to import.
    - ``verbose`` -- if ``True``, issue print statements when interesting examples are found
    - ``output`` -- a file path; if ``None``, overwrite the file given by ``location``

    EXAMPLES::

        sage: from sage.misc.replace_dot_all import *
        sage: import tempfile
        sage: with tempfile.TemporaryDirectory() as d:
        ....:     location = os.path.join(d, "input.py")
        ....:     with open(location, "w") as input:
        ....:         _ = input.write("from sage.plot.all import point2d\n")
        ....:         _ = input.write("from sage.plot.line import line\n")
        ....:     make_replacements_in_file(location, 'sage[.]plot[.]all', True)
        ....:     with open(location, "r") as output:
        ....:         for line in output:
        ....:             print(line.strip())
        from sage.plot.point import point2d
        from sage.plot.line import line
    """
    replacements = find_replacements(location, package_regex, verbose)
    with open(location) as file:
        lines = file.readlines()
    replaced_content = ""
    row_index = 0  # keeps track of the line number
    while row_index < len(lines):  # looping through the file
        line = lines[row_index]
        span = 0  # keeps track of number of lines import statement spans
        if replacements and row_index == replacements[0][0] and len(replacements[0]) == 4:
            span = replacements[0][3]  # if import statement spans span lines
        # returns the line if no replacements are needed and returns the processed line otherwise
        new_line, replacements = process_line(location, line, replacements, row_index, verbose=verbose)
        replaced_content += new_line + "\n"  # concatenate the new string and add an end-line break
        row_index += 1 + span
    if output is None:
        output = location
    with open(output, "w") as write_file:  # Open file in write mode
        write_file.write(replaced_content)  # overwriting the old file contents with the new/replaced content


def walkdir_replace_dot_all(dir, file_regex=r'.*[.](py|pyx|pxi)$', package_regex=None, verbose=False, *,
                            excluded_file_regex=r'auto-methods|replace_dot_all'):
    r"""
    Replace ``import`` statements in the files in directory ``dir`` matching the regex pattern ``file_regex``.

    INPUTS:

    - ``dir`` -- a directory path
    - ``file_regex`` -- a regular expression matching the file names to process
    - ``package_regex`` -- (default: :obj:`default_package_regex`) a regular expression matching
      the ``sage.PAC.KAGE.all`` package names from which we do not want to import.
    - ``verbose`` -- if ``True``, print statements when interesting examples are found
    - ``excluded_file_regex`` -- a regular expression matching the file names to exclude

    EXAMPLES::

        sage: # needs SAGE_SRC
        sage: from sage.misc.replace_dot_all import *
        sage: walkdir_replace_dot_all(os.path.join(sage.env.SAGE_SRC, 'sage'))  # not tested
    """
    global numberFiles, numberFilesMatchingRegex
    file_regex = re.compile(file_regex)
    excluded_file_regex = re.compile(excluded_file_regex)
    for root, dirs, files in os.walk(dir, topdown=False):
        for name in files:
            numberFiles += 1
            if file_regex.search(name) and not excluded_file_regex.search(name):
                numberFilesMatchingRegex += 1
                location = os.path.join(root, name)
                make_replacements_in_file(location, package_regex, verbose)


# ******************************************************** EXECUTES MAIN FUNCTION **********************************************************************
# this executes the main function in this file which writes over all import statements matching regex in files in specified location matching fileRegex:
if __name__ == "__main__":
    # Create argument parser
    parser = argparse.ArgumentParser()
    # Optional arguments
    parser.add_argument(
        "location",
        metavar='files or directories',
        nargs='*',
        help=("Names of source directories or source files. "
              "If none given, walks through all files in src/sage."),
        type=str)
    parser.add_argument(
        "-v", "--verbose",
        help="Increase output verbosity. Shows locations of any unusual cases of import statements and the corresponding changes.",
        action="store_true")    # Parse arguments
    args = parser.parse_args()
    verbosity = args.verbose
    # Declare regular expressions
    file_regex = r'.*[.](py|pyx|pxi)$'
    package_regex = None
    # Execute the main function based on the specified location and verbosity
    if not args.location:
        from sage.env import SAGE_SRC

        args.location = [os.path.join(SAGE_SRC, 'sage')]
    try:
        for location in args.location:
            if not (location.endswith('.py') or location.endswith('.pxi')):
                # Assume directory
                walkdir_replace_dot_all(location, file_regex, package_regex, verbose=verbosity)
            else:
                # make replacements in file specified by location argument
                make_replacements_in_file(location, package_regex, verbose=verbosity)
    finally:
        # Print report also when interrupted
        if verbosity:
            log_messages_split = sorted(log_messages.rstrip().split('\n'))
            for i, message in enumerate(log_messages_split, start=1):
                # add index to each line
                print(f'{i}. {message.rstrip()}')
        report = 'REPORT:\n'
        report += f'Number of files checked: {numberFiles}\n'
        report += f'Number of files matching regex: {numberFilesMatchingRegex}\n'
        report += f'Number of files changed: {numberFilesChanged}\n'
        report += f'Number of import statements replaced: {numberStatementsReplaced}'
        print('*'*100 + '\n' + report + '\n' + '*'*100)

# ******************************************************************************************************************************************************



# sage_setup: distribution = sagemath-symbolics
from sage.calculus.all__sagemath_gsl import *

from sage.calculus import desolvers
from sage.calculus.calculus import maxima as maxima_calculus
from sage.calculus.calculus import (laplace, inverse_laplace,
                                    limit, lim)

from sage.calculus.desolvers import (desolve, desolve_laplace, desolve_system,
                                     eulers_method, eulers_method_2x2,
                                     eulers_method_2x2_plot, desolve_rk4, desolve_system_rk4,
                                     desolve_odeint, desolve_mintides, desolve_tides_mpfr)
from sage.calculus.expr import symbolic_expression
from sage.calculus.var import (var, function, clear_vars)



# sage_setup: distribution = sagemath-gsl
from sage.misc.lazy_import import lazy_import

lazy_import("sage.calculus.transforms.fft", ["FastFourierTransform", "FFT"])
lazy_import("sage.calculus.transforms.dwt", ["WaveletTransform", "DWT"])
from sage.calculus.transforms.dft import IndexedSequence
del lazy_import



# sage_setup: distribution = sagemath-pari

from sage.libs.gmp.mpz cimport mpz_t

from cypari2.paridecl cimport (GEN, cgetg, t_POL, set_gel, gel, stoi, lg,
                               evalvarn, evalsigne, Z_issquare,
                               hyperellratpoints)
from cypari2.stack cimport clear_stack
from cysignals.signals cimport sig_on

from sage.libs.pari.convert_gmp cimport _new_GEN_from_mpz_t


cdef bint ratpoints_mpz_exists_only(mpz_t *coeffs, long degree, long H) except -1:
    """
    Search for projective points on the hyperelliptic curve
    ``y^2 = P(x)``.

    INPUT:

    - ``coeffs`` -- an array of length ``degree + 1`` giving the
      coefficients of ``P``, starting with the constant coefficient

    - ``degree`` -- degree of ``P``

    - ``H`` -- bound on the naive height for search

    OUTPUT: boolean, whether or not a projective point was found
    """
    sig_on()
    cdef GEN pol = cgetg(degree + 3, t_POL)
    pol[1] = evalvarn(0) + evalsigne(1)
    cdef long i
    for i in range(degree + 1):
        set_gel(pol, i+2, _new_GEN_from_mpz_t(coeffs[i]))

    # PARI checks only for affine points, so we manually check for
    # points at infinity (of the smooth model)
    cdef int r
    if degree % 2 or Z_issquare(gel(pol, degree+2)):
        r = 1
    else:
        R = hyperellratpoints(pol, stoi(H), 1)
        r = (lg(R) > 1)
    clear_stack()
    return r



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/calcium.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    const char * calcium_version() noexcept
    ulong calcium_fmpz_hash(const fmpz_t x) noexcept
    void calcium_stream_init_file(calcium_stream_t out, FILE * fp) noexcept
    void calcium_stream_init_str(calcium_stream_t out) noexcept
    void calcium_write(calcium_stream_t out, const char * s) noexcept
    void calcium_write_free(calcium_stream_t out, char * s) noexcept
    void calcium_write_si(calcium_stream_t out, slong x) noexcept
    void calcium_write_fmpz(calcium_stream_t out, const fmpz_t x) noexcept
    void calcium_write_arb(calcium_stream_t out, const arb_t z, slong digits, ulong flags) noexcept
    void calcium_write_acb(calcium_stream_t out, const acb_t z, slong digits, ulong flags) noexcept



# sage_setup: distribution = sagemath-schemes
# sage.doctest: needs sage.libs.pari
r"""
Hom spaces between Hecke modules
"""
# ****************************************************************************
#  Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.matrix.constructor import matrix
from sage.matrix.matrix_space import MatrixSpace
from sage.categories.homset import HomsetWithBase
from .morphism import HeckeModuleMorphism_matrix
from .module import HeckeModule_generic


class HeckeModuleHomspace(HomsetWithBase):
    r"""
    A space of homomorphisms between two objects in the category of Hecke
    modules over a given base ring.
    """
    def __init__(self, X, Y, category=None) -> None:
        r"""
        Create the space of homomorphisms between X and Y, which must have the
        same base ring.

        EXAMPLES::

            sage: M = ModularForms(Gamma0(7), 4)
            sage: M.Hom(M)
            Set of Morphisms from ... to ... in Category of Hecke modules over Rational Field
            sage: sage.modular.hecke.homspace.HeckeModuleHomspace(M, M.base_extend(Qp(13)))
            Traceback (most recent call last):
            ...
            TypeError: X and Y must have the same base ring
            sage: M.Hom(M) == loads(dumps(M.Hom(M)))
            True

        TESTS::

            sage: M = ModularForms(Gamma0(7), 4)
            sage: H = M.Hom(M)
            sage: TestSuite(H).run(skip='_test_elements')
        """
        if not isinstance(X, HeckeModule_generic) or not isinstance(Y, HeckeModule_generic):
            raise TypeError("X and Y must be Hecke modules")
        if X.base_ring() != Y.base_ring():
            raise TypeError("X and Y must have the same base ring")
        if category is None:
            category = X.category()
        HomsetWithBase.__init__(self, X, Y, category=category)

    def __call__(self, A, name='', **kwds):
        r"""
        Create an element of the homspace ``self`` from `A`.

        INPUT:

        - ``A`` -- one of the following:

          - an element of a Hecke algebra

          - a Hecke module morphism

          - a matrix

          - a list of elements of the codomain specifying the images
            of the basis elements of the domain.

        EXAMPLES::

            sage: M = ModularForms(Gamma0(7), 4)
            sage: H = M.Hom(M)
            sage: H(M.hecke_operator(7))
            Hecke module morphism T_7 defined by the matrix
            [ -7   0   0]
            [  0   1 240]
            [  0   0 343]
            Domain:   Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...
            Codomain: Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...
            sage: H(H(M.hecke_operator(7)))
            Hecke module morphism T_7 defined by the matrix
            [ -7   0   0]
            [  0   1 240]
            [  0   0 343]
            Domain:   Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...
            Codomain: Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...
            sage: H(matrix(QQ, 3, srange(9)))
            Hecke module morphism defined by the matrix
            [0 1 2]
            [3 4 5]
            [6 7 8]
            Domain:   Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...
            Codomain: Modular Forms space of dimension 3 for Congruence Subgroup Gamma0(7) ...

        TESTS:

        Make sure that the element is created correctly when the codomain is
        not the full module (related to :issue:`21497`)::

            sage: M = ModularSymbols(Gamma0(3), weight=22, sign=1)
            sage: S = M.cuspidal_subspace()
            sage: H = S.Hom(S)
            sage: H(S.gens())
            Hecke module morphism defined by the matrix
            [1 0 0 0 0 0]
            [0 1 0 0 0 0]
            [0 0 1 0 0 0]
            [0 0 0 1 0 0]
            [0 0 0 0 1 0]
            [0 0 0 0 0 1]
            Domain:   Modular Symbols subspace of dimension 6 of Modular Symbols space ...
            Codomain: Modular Symbols subspace of dimension 6 of Modular Symbols space ...

            sage: H.zero() in H
            True
            sage: H.one() in H
            True
        """
        try:
            if A.parent() == self:
                A._set_parent(self)
                return A
            A = A.hecke_module_morphism()
            if A.parent() == self:
                A._set_parent(self)
                return A
            else:
                raise TypeError("unable to coerce A to self")
        except AttributeError:
            pass
        side = kwds.get("side", "left")
        if A in self.base_ring():
            dim_dom = self.domain().rank()
            dim_codom = self.codomain().rank()
            if side == "left":
                MS = MatrixSpace(self.base_ring(), dim_dom, dim_codom)
            else:
                MS = MatrixSpace(self.base_ring(), dim_codom, dim_dom)
            if self.domain() == self.codomain():
                A = A * MS.identity_matrix()
            elif A == 0:
                A = MS.zero()
            else:
                raise ValueError('scalars do not coerce to this homspace')
        elif isinstance(A, (list, tuple)):
            A = matrix([self.codomain().coordinate_vector(f) for f in A])
        if side == "right":
            A = A.transpose()
        return HeckeModuleMorphism_matrix(self, A, name, side)

    def _an_element_(self):
        """
        Return an element.

        If the domain is equal to the codomain, this returns the
        action of the Hecke operator of index 2. Otherwise, this returns zero.

        EXAMPLES::

            sage: # needs sage.libs.flint
            sage: M = ModularSymbols(Gamma0(2), weight=12, sign=1)
            sage: S = M.cuspidal_subspace()
            sage: S.Hom(S).an_element()
            Hecke module morphism defined by the matrix
            [      260 -2108/135]
            [     4860      -284]
            Domain:   Modular Symbols subspace of dimension 2 of Modular Symbols space ...
            Codomain: Modular Symbols subspace of dimension 2 of Modular Symbols space ...
        """
        if self.domain() != self.codomain():
            return self.zero()
        else:
            A = self.domain().hecke_operator(2).matrix()
            return HeckeModuleMorphism_matrix(self, A)



# sage_setup: distribution = sagemath-flint
# Deprecated header file; use sage/libs/flint/types.pxd instead
# See https://github.com/sagemath/sage/pull/36449

from sage.libs.flint.types cimport (
    mag_struct,
    mag_t,
    mag_ptr,
    mag_srcptr,
    arf_struct,
    arf_t,
    arf_ptr,
    arf_srcptr,
    arf_rnd_t,
    ARF_RND_DOWN,
    ARF_RND_UP,
    ARF_RND_FLOOR,
    ARF_RND_CEIL,
    ARF_RND_NEAR,
    ARF_PREC_EXACT,
    arb_struct,
    arb_t,
    arb_ptr,
    arb_srcptr,
    acb_struct,
    acb_t,
    acb_ptr,
    acb_srcptr,
    acb_mat_struct,
    acb_mat_t,
    acb_poly_struct,
    acb_poly_t,
    acb_calc_integrate_opt_struct,
    acb_calc_integrate_opt_t,
    acb_calc_func_t,
    arb_poly_struct,
    arb_poly_t)



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.modules
r"""
Exterior algebras Gröbner bases

This contains the backend implementations in Cython for the Gröbner bases
of exterior algebra.

AUTHORS:

- Trevor K. Karn, Travis Scrimshaw (July 2022): Initial implementation
"""

# ****************************************************************************
#       Copyright (C) 2022 Trevor K. Karn <karnx018 at umn.edu>
#                 (C) 2022 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_check
from sage.libs.gmp.mpz cimport mpz_sizeinbase, mpz_setbit, mpz_tstbit, mpz_sgn
from sage.data_structures.bitset_base cimport (bitset_init, bitset_first,
                                               bitset_next, bitset_set_to, bitset_len)
from sage.structure.parent cimport Parent
from sage.structure.richcmp cimport richcmp, rich_to_bool
from sage.data_structures.blas_dict cimport iaxpy
from copy import copy

cdef inline long degree(FrozenBitset X) noexcept:
    """
    Compute the degree of ``X``.
    """
    return bitset_len(X._bitset)


cdef inline CliffordAlgebraElement build_monomial(Parent E, FrozenBitset supp):
    """
    Helper function for the fastest way to build a monomial.
    """
    return <CliffordAlgebraElement> E.element_class(E, {supp: E._base.one()})

cdef class GBElement:
    """
    Helper class for storing an element with its leading support both as
    a :class:`FrozenBitset` and an :class:`Integer`.
    """
    def __init__(self, CliffordAlgebraElement x, FrozenBitset ls, Integer n):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import GBElement
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: X = GBElement(a, a.leading_support(), 1)
            sage: TestSuite(X).run(skip='_test_pickling')
        """
        self.elt = x
        self.ls = ls
        self.lsi = n

    def __hash__(self):
        """
        Return the hash of ``self``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import GBElement
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: X = GBElement(a, a.leading_support(), 1)
            sage: hash(X) == 1
            True
        """
        return int(self.lsi)

    def __richcmp__(self, other, int op):
        """
        Rich compare ``self`` with ``other`` by ``op``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import GBElement
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: X = GBElement(a, a.leading_support(), 1)
            sage: Y = GBElement(a*b, (a*b).leading_support(), 3)
            sage: X == X
            True
            sage: X == Y
            False
            sage: X != Y
            True
        """
        if self is other:
            return rich_to_bool(op, 0)
        return richcmp(self.elt, (<GBElement> other).elt, op)

cdef class GroebnerStrategy:
    """
    A strategy for computing a Gröbner basis.

    INPUT:

    - ``I`` -- the ideal
    """
    def __init__(self, I):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import GroebnerStrategy
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([a + 1], side='left')
            sage: G = GroebnerStrategy(I)
            sage: TestSuite(G).run(skip='_test_pickling')
        """
        self.ideal = I
        self.groebner_basis = (None,)
        self.E = <Parent> I.ring()
        self.homogeneous = I._homogeneous
        self.rank = Integer(self.E.ngens())
        if self.homogeneous or I.side() == "left":
            self.side = 0
        elif I.side() == "right":
            self.side = 1
        else:
            self.side = 2

    cdef inline FrozenBitset leading_support(self, CliffordAlgebraElement f):
        """
        Return the leading support of the exterior algebra element ``f``.
        """
        cdef dict mc = <dict> f._monomial_coefficients
        return self.int_to_bitset(max(self.bitset_to_int(k) for k in mc))

    cdef inline partial_S_poly_left(self, GBElement f, GBElement g):
        r"""
        Compute one half of the `S`-polynomial for ``f`` and ``g``.

        This computes:

        .. MATH::

            LCM(LM(f), LM(g)) / LT(f) \cdot f.
        """
        cdef FrozenBitset D = <FrozenBitset> g.ls.difference(f.ls)
        cdef GBElement ret = self.prod_term_GB(D, f)
        inv = ~ret.elt[ret.ls]
        for k in ret.elt._monomial_coefficients:
            ret.elt._monomial_coefficients[k] *= inv
        return ret

    cdef inline partial_S_poly_right(self, GBElement f, GBElement g):
        r"""
        Compute one half of the `S`-polynomial for ``f`` and ``g``.

        This computes:

        .. MATH::

            f \cdot LCM(LM(f), LM(g)) / LT(f).
        """
        cdef FrozenBitset D = <FrozenBitset> g.ls.difference(f.ls)
        cdef GBElement ret = self.prod_GB_term(f, D)
        inv = ~ret.elt[ret.ls]
        for k in ret.elt._monomial_coefficients:
            ret.elt._monomial_coefficients[k] *= inv
        return ret

    cdef inline GBElement build_elt(self, CliffordAlgebraElement f):
        """
        Convert ``f`` into a ``GBElement``.
        """
        cdef dict mc = <dict> f._monomial_coefficients
        # if not mc:
        #     return GBElement(f, FrozenBitset(), -1)
        cdef Integer r = <Integer> max(self.bitset_to_int(k) for k in mc)
        return GBElement(f, self.int_to_bitset(r), r)

    cdef inline GBElement prod_GB_term(self, GBElement f, FrozenBitset t):
        """
        Return the GBElement corresponding to ``f * t``.

        .. WARNING::

            This assumes the leading support is ``f.ls._union(t)``.
        """
        ret = f.elt._mul_self_term(self.E, self.E._base.one())
        cdef FrozenBitset ls = <FrozenBitset> f.ls._union(t)
        return GBElement(<CliffordAlgebraElement> ret, ls, self.bitset_to_int(ls))

    cdef inline GBElement prod_term_GB(self, FrozenBitset t, GBElement f):
        """
        Return the GBElement corresponding to ``t * f``.

        .. WARNING::

            This assumes the leading support is ``f.ls._union(t)``.
        """
        ret = f.elt._mul_term_self(t, self.E._base.one())
        cdef FrozenBitset ls = <FrozenBitset> f.ls._union(t)
        return GBElement(<CliffordAlgebraElement> ret, ls, self.bitset_to_int(ls))

    cdef inline bint build_S_poly(self, GBElement f, GBElement g) noexcept:
        r"""
        Check to see if we should build the `S`-polynomial.

        For homogeneous ideals, we throw out all those pairs `(f, g)` such that

        .. MATH::

            LCM(LM(f), LM(g)) == LM(f) \cdot LM(g).
        """
        if not self.homogeneous:
            return True

        return (<FrozenBitset> f.ls.intersection(g.ls)).isempty()

    cdef inline set preprocessing(self, list P, list G):
        """
        Perform the preprocessing step.
        """
        cdef GBElement f, g, f0, f1
        cdef set additions

        cdef set L = set()
        if self.side == 1:
            for f0, f1 in P:
                if self.build_S_poly(f0, f1):
                    L.add(self.partial_S_poly_right(f0, f1))
                    L.add(self.partial_S_poly_right(f1, f0))
        else:  # We compute a left Gröbner basis for two-sided ideals
            for f0, f1 in P:
                if self.build_S_poly(f0, f1):
                    L.add(self.partial_S_poly_left(f0, f1))
                    L.add(self.partial_S_poly_left(f1, f0))

        if self.side == 2 and not self.homogeneous:
            # Add in all S-poly times positive degree monomials
            one = self.E._base.one()
            additions = set((<GBElement> f).elt._mul_self_term(t, one) for t in self.E._indices for f in L)
            L.update(self.build_elt(f) for f in additions if f)

        cdef set done = set((<GBElement> f).ls for f in L)
        cdef set monL = set()
        for f in L:
            monL.update(f.elt._monomial_coefficients)
        monL.difference_update(done)

        while monL:
            m = self.int_to_bitset(max(self.bitset_to_int(k) for k in monL))
            done.add(m)
            monL.remove(m)
            for g in G:
                lm = g.ls
                if lm <= m:
                    f = self.prod_term_GB(<FrozenBitset> m.difference(lm), g)
                    if f in L:
                        break
                    monL.update(set(f.elt._monomial_coefficients) - done)
                    L.add(f)
                    break
        return L

    cdef inline list reduction(self, list P, list G):
        """
        Perform the reduction of ``P`` mod ``G`` in ``E``.
        """
        cdef set L = self.preprocessing(P, G)
        cdef Py_ssize_t i
        from sage.matrix.constructor import matrix
        cdef Integer r = Integer(2) ** self.rank - Integer(1)  # r for "rank" or "reverso"
        M = matrix({(i, r - self.bitset_to_int(<FrozenBitset> m)): c
                    for i, f in enumerate(L)
                    for m, c in (<GBElement> f).elt._monomial_coefficients.items()},
                   sparse=True)
        M.echelonize()  # Do this in place
        lead_supports = set((<GBElement> f).lsi for f in L)
        return [GBElement(self.E.element_class(self.E, {self.int_to_bitset(r - Integer(j)): c for j, c in M[i].items()}),
                          self.int_to_bitset(Integer(r - p)),
                          Integer(r - p))
                for i, p in enumerate(M.pivots())
                if r - Integer(p) not in lead_supports]

    def compute_groebner(self, reduced=True):
        r"""
        Compute the (``reduced``) left/right Gröbner basis for the ideal ``I``.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: E.<y, x> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([x*y - x, x*y - 1], side='left')
            sage: I.groebner_basis()  # indirect doctest
            (1,)
            sage: J = E.ideal([x*y - x, 2*x*y - 2], side='left')
            sage: J.groebner_basis()  # indirect doctest
            (1,)

            sage: # needs sage.libs.pari
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([a+b*c], side='left')
            sage: I.groebner_basis()  # indirect doctest
            (b*c + a,)
            sage: I = E.ideal([a+b*c], side='twosided')
            sage: I.groebner_basis()  # indirect doctest
            (a*b, a*c, b*c + a, a*d)
        """
        cdef FrozenBitset p0, p1
        cdef long deg
        cdef Py_ssize_t i, j
        cdef GBElement f0, f1
        cdef list G = [], Gp
        cdef dict constructed = {}
        cdef CliffordAlgebraElement f

        for f in self.ideal.gens():
            if not f:  # Remove 0s
                continue
            f0 = self.build_elt(f)
            if f0.lsi in constructed:
                if f0 in constructed[f0.lsi]:  # Already there
                    continue
                constructed[f0.lsi].add(f0)
            else:
                constructed[f0.lsi] = set([f0])
            G.append(f0)

        if self.side == 2 and not self.homogeneous:
            # Add in all S-poly times positive degree monomials
            one = self.E._base.one()
            for t in self.E._indices:
                for f0 in G:
                    f = f0.elt._mul_self_term(t, one)
                    if not f:
                        continue
                    f1 = self.build_elt(f)
                    if f1.lsi in constructed:
                        if f1 in constructed[f1.lsi]:  # Already there
                            continue
                        constructed[f1.lsi].add(f1)
                    else:
                        constructed[f1.lsi] = set([f1])
                    G.append(f1)

        cdef Py_ssize_t n = len(G)
        cdef dict P = {}

        for i in range(n):
            f0 = <GBElement> G[i]
            p0 = f0.ls
            for j in range(i+1, n):
                f1 = <GBElement> G[j]
                p1 = f1.ls
                deg = degree(<FrozenBitset> (p0._union(p1)))
                if deg in P:
                    P[deg].append((f0, f1))
                else:
                    P[deg] = [(f0, f1)]

        while P:
            sig_check()
            Pp = P.pop(min(P))  # The selection: lowest lcm degree
            Gp = self.reduction(Pp, G)
            # Add the elements Gp to G when a new element is found
            for f0 in Gp:
                if f0.lsi in constructed:
                    if f0 in constructed[f0.lsi]:  # Already there
                        continue
                    constructed[f0.lsi].add(f0)
                else:
                    constructed[f0.lsi] = set([f0])
                G.append(f0)
            # Find the degrees of the new pairs
            for j in range(n, len(G)):
                f1 = G[j]
                p1 = f1.ls
                for i in range(j):
                    f0 = G[i]
                    p0 = f0.ls
                    deg = degree(<FrozenBitset> (p0._union(p1)))
                    if deg in P:
                        P[deg].append((f0, f1))
                    else:
                        P[deg] = [(f0, f1)]
            n = len(G)

        G.sort(key=lambda x: (<GBElement> x).lsi)

        if not reduced:
            self.groebner_basis = tuple([(<GBElement> f0).elt for f0 in G if (<GBElement> f0).elt])
            return
        self.reduced_gb(G)

    cdef int reduced_gb(self, list G) except -1:
        """
        Convert the Gröbner basis ``G`` into a reduced Gröbner basis.
        """
        cdef Py_ssize_t i, j, k
        cdef GBElement f0, f1

        # Now that we have a Gröbner basis, we make this into a reduced Gröbner basis
        cdef Py_ssize_t num_zeros = 0
        cdef Py_ssize_t n = len(G)
        cdef set pairs = set((i, j) for i in range(n) for j in range(n) if i != j)

        while pairs:
            sig_check()
            i, j = pairs.pop()
            f0 = <GBElement> G[i]
            f1 = <GBElement> G[j]
            assert f0.elt._monomial_coefficients is not f1.elt._monomial_coefficients, (i, j)
            # We perform the classical reduction algorithm here on each pair
            # TODO: Make this faster by using the previous technique?
            if self.reduce_single(f0.elt, f1.elt):
                if f0.elt:
                    G[i] = self.build_elt(f0.elt)
                    pairs.update((k, i) for k in range(n) if k != i)
                else:
                    G[i] = GBElement(f0.elt, FrozenBitset(), Integer(2)**self.rank + 1)
                    num_zeros += 1
                    pairs.difference_update((k, i) for k in range(n) if k != i)
                    pairs.difference_update((i, k) for k in range(n) if k != i)

        G.sort(key=lambda x: (<GBElement> x).lsi)
        for i in range(len(G)-num_zeros):
            f0 = <GBElement> G[i]
            if f0.elt:
                inv = ~f0.elt[f0.ls]
                for key in f0.elt._monomial_coefficients:
                    f0.elt._monomial_coefficients[key] *= inv
        self.groebner_basis = tuple([f0.elt for f0 in G[:len(G)-num_zeros]])
        return 0

    def reduce_computed_gb(self):
        """
        Convert the computed Gröbner basis to a reduced Gröbner basis.

        EXAMPLES::

            sage: # needs sage.libs.pari
            sage: E.<x,y,z> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([x+y*z])
            sage: I.groebner_basis(reduced=False)
            (x*y, x*z, y*z + x, x*y*z)
            sage: I._groebner_strategy.reduce_computed_gb()
            sage: I._groebner_strategy.groebner_basis
            (x*y, x*z, y*z + x)
        """
        if self.groebner_basis == [(None,)]:
            raise ValueError("Gröbner basis not yet computed")
        cdef list G = [self.build_elt(f) for f in self.groebner_basis]
        self.reduced_gb(G)

    cpdef CliffordAlgebraElement reduce(self, CliffordAlgebraElement f):
        """
        Reduce ``f`` modulo the ideal with Gröbner basis ``G``.

        EXAMPLES::

            sage: E.<a,b,c,d,e> = ExteriorAlgebra(QQ)
            sage: rels = [c*d*e - b*d*e + b*c*e - b*c*d,
            ....:         c*d*e - a*d*e + a*c*e - a*c*d,
            ....:         b*d*e - a*d*e + a*b*e - a*b*d,
            ....:         b*c*e - a*c*e + a*b*e - a*b*c,
            ....:         b*c*d - a*c*d + a*b*d - a*b*c]
            sage: I = E.ideal(rels)
            sage: GB = I.groebner_basis()
            sage: I._groebner_strategy.reduce(a*b*e)
            a*b*e
            sage: I._groebner_strategy.reduce(b*d*e)
            a*b*d - a*b*e + a*d*e
            sage: I._groebner_strategy.reduce(c*d*e)
            a*c*d - a*c*e + a*d*e
            sage: I._groebner_strategy.reduce(a*b*c*d*e)
            0
            sage: I._groebner_strategy.reduce(a*b*c*d)
            0
            sage: I._groebner_strategy.reduce(E.zero())
            0

        Check :issue:`37108` is fixed::

            sage: # needs sage.libs.pari
            sage: E = ExteriorAlgebra(QQ, 6)
            sage: E.inject_variables(verbose=False)
            sage: gens = [-e0*e1*e2 + e0*e1*e5 - e0*e2*e3 - e0*e3*e5 + e1*e2*e3 + e1*e3*e5,
            ....:         e1*e2 - e1*e5 + e2*e5, e0*e2 - e0*e4 + e2*e4,
            ....:         e3*e4 - e3*e5 + e4*e5, e0*e1 - e0*e3 + e1*e3]
            sage: I = E.ideal(gens)
            sage: S = E.quo(I)
            sage: I.reduce(e1*e3*e4*e5)
            0
        """
        if not f:
            return f
        # Make a copy to mutate
        f = type(f)(f._parent, copy(f._monomial_coefficients))
        was_reduced = True
        while was_reduced:
            was_reduced = False
            for g in self.groebner_basis:
                was_reduced = self.reduce_single(f, g)
                if was_reduced:
                    break
        return f

    cdef bint reduce_single(self, CliffordAlgebraElement f, CliffordAlgebraElement g) except -1:
        r"""
        Reduce ``f`` by ``g``.

        .. WARNING::

            This modifies the element ``f``.
        """
        cdef FrozenBitset lm = self.leading_support(g), s, t
        cdef bint did_reduction = True, was_reduced=False
        cdef CliffordAlgebraElement gp

        one = self.E._base.one()
        while did_reduction:
            did_reduction = False
            for s in f._monomial_coefficients:
                if lm.issubset(s):
                    t = s
                    did_reduction = True
                    was_reduced = True
                    break
            if did_reduction:
                if self.side == 0:
                    gp = g._mul_term_self(t - lm, one)
                else:
                    gp = g._mul_self_term(t - lm, one)
                coeff = f[t] / gp._monomial_coefficients[t]
                iaxpy(-coeff, gp._monomial_coefficients, f._monomial_coefficients)
        return was_reduced

    cdef Integer bitset_to_int(self, FrozenBitset X):
        raise NotImplementedError

    cdef FrozenBitset int_to_bitset(self, Integer n):
        raise NotImplementedError

    def sorted_monomials(self, as_dict=False):
        """
        Helper function to display the monomials in their term order
        used by ``self``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import *
            sage: E.<x,y,z> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([x, y])
            sage: GroebnerStrategyNegLex(I).sorted_monomials()
            [1, x, y, x*y, z, x*z, y*z, x*y*z]
            sage: GroebnerStrategyDegLex(I).sorted_monomials()
            [1, x, y, z, x*y, x*z, y*z, x*y*z]
            sage: GroebnerStrategyDegRevLex(I).sorted_monomials()
            [1, z, y, x, y*z, x*z, x*y, x*y*z]

            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([a, b])
            sage: GroebnerStrategyNegLex(I).sorted_monomials()
            [1,
             a,
             b, a*b,
             c, a*c, b*c, a*b*c,
             d, a*d, b*d, a*b*d, c*d, a*c*d, b*c*d, a*b*c*d]
            sage: GroebnerStrategyDegLex(I).sorted_monomials()
            [1,
             a, b, c, d,
             a*b, a*c, a*d, b*c, b*d, c*d,
             a*b*c, a*b*d, a*c*d, b*c*d,
             a*b*c*d]
            sage: GroebnerStrategyDegRevLex(I).sorted_monomials()
            [1,
             d, c, b, a,
             c*d, b*d, a*d, b*c, a*c, a*b,
             b*c*d, a*c*d, a*b*d, a*b*c,
             a*b*c*d]
        """
        cdef FrozenBitset X
        cdef Integer i
        cdef list D = [self.bitset_to_int(X) for X in self.E._indices]
        D.sort()
        if as_dict:
            return {i: build_monomial(self.E, self.int_to_bitset(i)) for i in D}
        return [build_monomial(self.E, self.int_to_bitset(i)) for i in D]

    def monomial_to_int(self):
        """
        Helper function to display the monomials in their term order
        used by ``self``.

        EXAMPLES::

            sage: from sage.algebras.exterior_algebra_groebner import *
            sage: E.<a,b,c,d> = ExteriorAlgebra(QQ)
            sage: I = E.ideal([a, b])
            sage: GroebnerStrategyDegLex(I).monomial_to_int()
            {1: 0,
             a: 1, b: 2, c: 3, d: 4,
             a*b: 5, a*c: 6, a*d: 7, b*c: 8, b*d: 9, c*d: 10,
             a*b*c: 11, a*b*d: 12, a*c*d: 13, b*c*d: 14,
             a*b*c*d: 15}
            sage: GroebnerStrategyDegRevLex(I).monomial_to_int()
            {1: 0,
             a: 4, b: 3, c: 2, d: 1,
             a*b: 10, a*c: 9, a*d: 7, b*c: 8, b*d: 6, c*d: 5,
             a*b*c: 14, a*b*d: 13, a*c*d: 12, b*c*d: 11,
             a*b*c*d: 15}
        """
        B = self.E.basis()
        return {B[X]: self.bitset_to_int(X) for X in self.E._indices}

cdef class GroebnerStrategyNegLex(GroebnerStrategy):
    """
    Gröbner basis strategy implementing neglex ordering.
    """
    cdef inline Integer bitset_to_int(self, FrozenBitset X):
        """
        Convert ``X`` to an :class:`Integer`.
        """
        cdef Integer ret = Integer(0)
        cdef long elt = bitset_first(X._bitset)
        while elt >= 0:
            mpz_setbit(ret.value, elt)
            elt = bitset_next(X._bitset, elt + 1)
        return ret

    cdef inline FrozenBitset int_to_bitset(self, Integer n):
        """
        Convert a nonnegative integer ``n`` to a :class:`FrozenBitset`.
        """
        cdef size_t i

        if mpz_sgn(n.value) == 0:
            return FrozenBitset()

        cdef FrozenBitset ret = <FrozenBitset> FrozenBitset()
        cdef size_t s = mpz_sizeinbase(n.value, 2)
        bitset_init(ret._bitset, s)
        for i in range(s):
            bitset_set_to(ret._bitset, i, mpz_tstbit(n.value, i))
        return ret

cdef class GroebnerStrategyDegRevLex(GroebnerStrategy):
    """
    Gröbner basis strategy implementing degree revlex ordering.
    """
    cdef inline Integer bitset_to_int(self, FrozenBitset X):
        """
        Convert ``X`` to an :class:`Integer`.
        """
        if X.isempty():
            return Integer(0)

        cdef Integer n = self.rank
        cdef long i, deg = degree(X)
        cdef long r = 1
        cdef Integer t = Integer(0)

        cdef long elt = bitset_first(X._bitset)
        while elt >= 0:
            t += Integer(elt).binomial(r)
            r += 1
            elt = bitset_next(X._bitset, elt + 1)
        return Integer(sum(n.binomial(i) for i in range(deg+1)) - t - 1)

    cdef inline FrozenBitset int_to_bitset(self, Integer n):
        """
        Convert a nonnegative integer ``n`` to a :class:`FrozenBitset`.
        """
        if mpz_sgn(n.value) == 0:
            return FrozenBitset()

        cdef Py_ssize_t deg = 0
        cdef Integer binom = Integer(1)
        while n >= binom:
            n -= binom
            deg += 1
            binom = self.rank.binomial(deg)

        # TODO: Cythonize the from_rank
        from sage.combinat.combination import from_rank
        return FrozenBitset([self.rank - val - 1 for val in from_rank(n, self.rank, deg)])

cdef class GroebnerStrategyDegLex(GroebnerStrategy):
    """
    Gröbner basis strategy implementing degree lex ordering.
    """
    cdef inline Integer bitset_to_int(self, FrozenBitset X):
        """
        Convert ``X`` to an :class:`Integer`.
        """
        if X.isempty():
            return Integer(0)

        cdef Integer n = self.rank
        cdef long i, deg = degree(X)
        cdef long r = deg
        cdef Integer t = Integer(0)

        cdef long elt = bitset_first(X._bitset)
        while elt >= 0:
            t += Integer(n - 1 - elt).binomial(r)
            r -= 1
            elt = bitset_next(X._bitset, elt + 1)
        return Integer(sum(n.binomial(i) for i in range(deg+1)) - t - 1)

    cdef inline FrozenBitset int_to_bitset(self, Integer n):
        """
        Convert a nonnegative integer ``n`` to a :class:`FrozenBitset`.
        """
        if mpz_sgn(n.value) == 0:
            return FrozenBitset()

        cdef Py_ssize_t deg = 0
        cdef Integer binom = Integer(1)
        while n >= binom:
            n -= binom
            deg += 1
            binom = self.rank.binomial(deg)

        from sage.combinat.combination import from_rank
        return FrozenBitset(from_rank(n, self.rank, deg))



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/fmpz_factor.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fmpz_factor_init(fmpz_factor_t factor) noexcept
    void fmpz_factor_clear(fmpz_factor_t factor) noexcept
    void _fmpz_factor_append_ui(fmpz_factor_t factor, mp_limb_t p, ulong exp) noexcept
    void _fmpz_factor_append(fmpz_factor_t factor, const fmpz_t p, ulong exp) noexcept
    void fmpz_factor(fmpz_factor_t factor, const fmpz_t n) noexcept
    int fmpz_factor_smooth(fmpz_factor_t factor, const fmpz_t n, slong bits, int proved) noexcept
    void fmpz_factor_si(fmpz_factor_t factor, slong n) noexcept
    int fmpz_factor_trial_range(fmpz_factor_t factor, const fmpz_t n, ulong start, ulong num_primes) noexcept
    int fmpz_factor_trial(fmpz_factor_t factor, const fmpz_t n, slong num_primes) noexcept
    void fmpz_factor_refine(fmpz_factor_t res, const fmpz_factor_t f) noexcept
    void fmpz_factor_expand_iterative(fmpz_t n, const fmpz_factor_t factor) noexcept
    int fmpz_factor_pp1(fmpz_t factor, const fmpz_t n, ulong B1, ulong B2_sqrt, ulong c) noexcept
    int fmpz_factor_pollard_brent_single(fmpz_t p_factor, fmpz_t n_in, fmpz_t yi, fmpz_t ai, mp_limb_t max_iters) noexcept
    int fmpz_factor_pollard_brent(fmpz_t factor, flint_rand_t state, fmpz_t n, mp_limb_t max_tries, mp_limb_t max_iters) noexcept
    void fmpz_factor_ecm_init(ecm_t ecm_inf, mp_limb_t sz) noexcept
    void fmpz_factor_ecm_clear(ecm_t ecm_inf) noexcept
    void fmpz_factor_ecm_addmod(mp_ptr a, mp_ptr b, mp_ptr c, mp_ptr n, mp_limb_t n_size) noexcept
    void fmpz_factor_ecm_submod(mp_ptr x, mp_ptr a, mp_ptr b, mp_ptr n, mp_limb_t n_size) noexcept
    void fmpz_factor_ecm_double(mp_ptr x, mp_ptr z, mp_ptr x0, mp_ptr z0, mp_ptr n, ecm_t ecm_inf) noexcept
    void fmpz_factor_ecm_add(mp_ptr x, mp_ptr z, mp_ptr x1, mp_ptr z1, mp_ptr x2, mp_ptr z2, mp_ptr x0, mp_ptr z0, mp_ptr n, ecm_t ecm_inf) noexcept
    void fmpz_factor_ecm_mul_montgomery_ladder(mp_ptr x, mp_ptr z, mp_ptr x0, mp_ptr z0, mp_limb_t k, mp_ptr n, ecm_t ecm_inf) noexcept
    int fmpz_factor_ecm_select_curve(mp_ptr f, mp_ptr sigma, mp_ptr n, ecm_t ecm_inf) noexcept
    int fmpz_factor_ecm_stage_I(mp_ptr f, const mp_limb_t * prime_array, mp_limb_t num, mp_limb_t B1, mp_ptr n, ecm_t ecm_inf) noexcept
    int fmpz_factor_ecm_stage_II(mp_ptr f, mp_limb_t B1, mp_limb_t B2, mp_limb_t P, mp_ptr n, ecm_t ecm_inf) noexcept
    int fmpz_factor_ecm(fmpz_t f, mp_limb_t curves, mp_limb_t B1, mp_limb_t B2, flint_rand_t state, const fmpz_t n_in) noexcept



# sage_setup: distribution = sagemath-flint
# distutils: libraries = flint
# distutils: depends = flint/ca_ext.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/src/sage_setup/autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void ca_ext_init_qqbar(ca_ext_t res, const qqbar_t x, ca_ctx_t ctx) noexcept
    void ca_ext_init_const(ca_ext_t res, calcium_func_code func, ca_ctx_t ctx) noexcept
    void ca_ext_init_fx(ca_ext_t res, calcium_func_code func, const ca_t x, ca_ctx_t ctx) noexcept
    void ca_ext_init_fxy(ca_ext_t res, calcium_func_code func, const ca_t x, const ca_t y, ca_ctx_t ctx) noexcept
    void ca_ext_init_fxn(ca_ext_t res, calcium_func_code func, ca_srcptr x, slong nargs, ca_ctx_t ctx) noexcept
    void ca_ext_init_set(ca_ext_t res, const ca_ext_t x, ca_ctx_t ctx) noexcept
    void ca_ext_clear(ca_ext_t res, ca_ctx_t ctx) noexcept
    slong ca_ext_nargs(const ca_ext_t x, ca_ctx_t ctx) noexcept
    void ca_ext_get_arg(ca_t res, const ca_ext_t x, slong i, ca_ctx_t ctx) noexcept
    ulong ca_ext_hash(const ca_ext_t x, ca_ctx_t ctx) noexcept
    bint ca_ext_equal_repr(const ca_ext_t x, const ca_ext_t y, ca_ctx_t ctx) noexcept
    int ca_ext_cmp_repr(const ca_ext_t x, const ca_ext_t y, ca_ctx_t ctx) noexcept
    void ca_ext_print(const ca_ext_t x, ca_ctx_t ctx) noexcept
    void ca_ext_get_acb_raw(acb_t res, ca_ext_t x, slong prec, ca_ctx_t ctx) noexcept
    void ca_ext_cache_init(ca_ext_cache_t cache, ca_ctx_t ctx) noexcept
    void ca_ext_cache_clear(ca_ext_cache_t cache, ca_ctx_t ctx) noexcept
    ca_ext_ptr ca_ext_cache_insert(ca_ext_cache_t cache, const ca_ext_t x, ca_ctx_t ctx) noexcept



# sage_setup: distribution = sagemath-schemes
"""
Hyperelliptic curves of genus 2 over a general ring
"""
# ****************************************************************************
#  Copyright (C) 2006 David Kohel <kohel@maths.usyd.edu>
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from . import hyperelliptic_generic
from . import jacobian_g2
from . import invariants


class HyperellipticCurve_g2(hyperelliptic_generic.HyperellipticCurve_generic):
    def is_odd_degree(self):
        """
        Return ``True`` if the curve is an odd degree model.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x^4 + 3
            sage: HyperellipticCurve(f).is_odd_degree()
            True
        """
        f, h = self.hyperelliptic_polynomials()
        df = f.degree()
        if h.degree() < 3:
            return df % 2 == 1
        elif df < 6:
            return False

        a0 = f.leading_coefficient()
        c0 = h.leading_coefficient()
        return (c0**2 + 4*a0) == 0

    def jacobian(self):
        """
        Return the Jacobian of the hyperelliptic curve.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x^4 + 3
            sage: HyperellipticCurve(f).jacobian()
            Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^5 - x^4 + 3

        TESTS:

        Ensure that :issue:`37612` is fixed::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x^4 + 3
            sage: type(HyperellipticCurve(f).jacobian())
            <class 'sage.schemes.hyperelliptic_curves.jacobian_g2.HyperellipticJacobian_g2_with_category'>
        """
        return jacobian_g2.HyperellipticJacobian_g2(self)

    def kummer_morphism(self):
        """
        Return the morphism of an odd degree hyperelliptic curve to the Kummer
        surface of its Jacobian.

        This could be extended to an even degree model
        if a prescribed embedding in its Jacobian is fixed.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x^4 + 3
            sage: HyperellipticCurve(f).kummer_morphism()  # not tested
        """
        try:
            return self._kummer_morphism
        except AttributeError:
            pass
        if not self.is_odd_degree():
            raise TypeError("Kummer embedding not determined for even degree model curves.")
        self.jacobian().kummer_surface()
        return self._kummer_morphism

    def clebsch_invariants(self):
        r"""
        Return the Clebsch invariants `(A, B, C, D)` of Mestre, p 317, [Mes1991]_.

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.invariants`

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x^4 + 3
            sage: HyperellipticCurve(f).clebsch_invariants()
            (0, -2048/375, -4096/25, -4881645568/84375)
            sage: HyperellipticCurve(f(2*x)).clebsch_invariants()
            (0, -8388608/375, -1073741824/25, -5241627016305836032/84375)

            sage: HyperellipticCurve(f, x).clebsch_invariants()
            (-8/15, 17504/5625, -23162896/140625, -420832861216768/7119140625)
            sage: HyperellipticCurve(f(2*x), 2*x).clebsch_invariants()
            (-512/15, 71696384/5625, -6072014209024/140625, -451865844002031331704832/7119140625)

        TESTS::

            sage: # optional - magma
            sage: magma(HyperellipticCurve(f)).ClebschInvariants()
            [ 0, -2048/375, -4096/25, -4881645568/84375 ]
            sage: magma(HyperellipticCurve(f(2*x))).ClebschInvariants()
            [ 0, -8388608/375, -1073741824/25, -5241627016305836032/84375 ]
            sage: magma(HyperellipticCurve(f, x)).ClebschInvariants()
            [ -8/15, 17504/5625, -23162896/140625, -420832861216768/7119140625 ]
            sage: magma(HyperellipticCurve(f(2*x), 2*x)).ClebschInvariants()
            [ -512/15, 71696384/5625, -6072014209024/140625, -451865844002031331704832/7119140625 ]
        """
        f, h = self.hyperelliptic_polynomials()
        return invariants.clebsch_invariants(4*f + h**2)

    def igusa_clebsch_invariants(self):
        r"""
        Return the Igusa-Clebsch invariants `I_2, I_4, I_6, I_{10}` of Igusa and Clebsch [IJ1960]_.

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.invariants`

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: f = x^5 - x + 2
            sage: HyperellipticCurve(f).igusa_clebsch_invariants()
            (-640, -20480, 1310720, 52160364544)
            sage: HyperellipticCurve(f(2*x)).igusa_clebsch_invariants()
            (-40960, -83886080, 343597383680, 56006764965979488256)

            sage: HyperellipticCurve(f, x).igusa_clebsch_invariants()
            (-640, 17920, -1966656, 52409511936)
            sage: HyperellipticCurve(f(2*x), 2*x).igusa_clebsch_invariants()
            (-40960, 73400320, -515547070464, 56274284941110411264)

        TESTS::

            sage: # optional - magma
            sage: magma(HyperellipticCurve(f)).IgusaClebschInvariants()
            [ -640, -20480, 1310720, 52160364544 ]
            sage: magma(HyperellipticCurve(f(2*x))).IgusaClebschInvariants()
            [ -40960, -83886080, 343597383680, 56006764965979488256 ]
            sage: magma(HyperellipticCurve(f, x)).IgusaClebschInvariants()
            [ -640, 17920, -1966656, 52409511936 ]
            sage: magma(HyperellipticCurve(f(2*x), 2*x)).IgusaClebschInvariants()
            [ -40960, 73400320, -515547070464, 56274284941110411264 ]
        """
        f, h = self.hyperelliptic_polynomials()
        return invariants.igusa_clebsch_invariants(4*f + h**2)

    def absolute_igusa_invariants_wamelen(self):
        r"""
        Return the three absolute Igusa invariants used by van Wamelen [Wam1999]_.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: HyperellipticCurve(x^5 - 1).absolute_igusa_invariants_wamelen()
            (0, 0, 0)
            sage: HyperellipticCurve((x^5 - 1)(x - 2), (x^2)(x - 2)).absolute_igusa_invariants_wamelen()
            (0, 0, 0)
        """
        f, h = self.hyperelliptic_polynomials()
        return invariants.absolute_igusa_invariants_wamelen(4*f + h**2)

    def absolute_igusa_invariants_kohel(self):
        r"""
        Return the three absolute Igusa invariants used by Kohel [KohECHIDNA]_.

        .. SEEALSO::

            :meth:`sage.schemes.hyperelliptic_curves.invariants`

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: HyperellipticCurve(x^5 - 1).absolute_igusa_invariants_kohel()
            (0, 0, 0)
            sage: HyperellipticCurve(x^5 - x + 1, x^2).absolute_igusa_invariants_kohel()
            (-1030567/178769, 259686400/178769, 20806400/178769)
            sage: HyperellipticCurve((x^5 - x + 1)(3*x + 1), (x^2)(3*x + 1)).absolute_igusa_invariants_kohel()
            (-1030567/178769, 259686400/178769, 20806400/178769)
        """
        f, h = self.hyperelliptic_polynomials()
        return invariants.absolute_igusa_invariants_kohel(4*f + h**2)



# sage_setup: distribution = sagemath-modules
from sage.libs.gsl.types cimport *

from sage.libs.gsl.math cimport *
from sage.libs.gsl.complex cimport *
from sage.libs.gsl.poly cimport *
from sage.libs.gsl.airy cimport *
from sage.libs.gsl.bessel cimport *
from sage.libs.gsl.clausen cimport *
from sage.libs.gsl.coulomb cimport *
from sage.libs.gsl.coupling cimport *
from sage.libs.gsl.dawson cimport *
from sage.libs.gsl.debye cimport *
from sage.libs.gsl.dilog cimport *
from sage.libs.gsl.elementary cimport *
from sage.libs.gsl.ellint cimport *
from sage.libs.gsl.elljac cimport *
from sage.libs.gsl.erf cimport *
from sage.libs.gsl.exp cimport *
from sage.libs.gsl.expint cimport *
from sage.libs.gsl.fermi_dirac cimport *
from sage.libs.gsl.gamma cimport *
from sage.libs.gsl.gegenbauer cimport *
from sage.libs.gsl.hyperg cimport *
from sage.libs.gsl.laguerre cimport *
from sage.libs.gsl.lambert cimport *
from sage.libs.gsl.legendre cimport *
from sage.libs.gsl.log cimport *
from sage.libs.gsl.pow_int cimport *
from sage.libs.gsl.psi cimport *
from sage.libs.gsl.synchrotron cimport *
from sage.libs.gsl.transport cimport *
from sage.libs.gsl.trig cimport *
from sage.libs.gsl.wavelet cimport *
from sage.libs.gsl.zeta cimport *

from sage.libs.gsl.block cimport *
from sage.libs.gsl.vector cimport *
from sage.libs.gsl.vector_complex cimport *
from sage.libs.gsl.matrix cimport *
from sage.libs.gsl.matrix_complex cimport *

from sage.libs.gsl.permutation cimport *
from sage.libs.gsl.combination cimport *
from sage.libs.gsl.sort cimport *

from sage.libs.gsl.blas cimport *
from sage.libs.gsl.linalg cimport *
from sage.libs.gsl.eigen cimport *
from sage.libs.gsl.fft cimport *
from sage.libs.gsl.integration cimport *
from sage.libs.gsl.rng cimport *
from sage.libs.gsl.qrng cimport *
from sage.libs.gsl.random cimport *
from sage.libs.gsl.statistics cimport *
from sage.libs.gsl.histogram cimport *
from sage.libs.gsl.ntuple cimport *
from sage.libs.gsl.monte cimport *
from sage.libs.gsl.odeiv cimport *
from sage.libs.gsl.interp cimport *
from sage.libs.gsl.chebyshev cimport *
from sage.libs.gsl.sum cimport *
from sage.libs.gsl.roots cimport *
from sage.libs.gsl.min cimport *
from sage.libs.gsl.fit cimport *
from sage.libs.gsl.errno cimport *



# sage_setup: distribution = sagemath-gap
r"""
Permutation group homomorphisms

AUTHORS:

- David Joyner (2006-03-21): first version

- David Joyner (2008-06): fixed kernel and image to return a group,
  instead of a string.

EXAMPLES::

    sage: G = CyclicPermutationGroup(4)
    sage: H = DihedralGroup(4)
    sage: g = G([(1,2,3,4)])
    sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens()))
    sage: phi.image(G)
    Subgroup generated by [(1,2,3,4)] of
     (Dihedral group of order 8 as a permutation group)
    sage: phi.kernel()
    Subgroup generated by [()] of (Cyclic group of order 4 as a permutation group)
    sage: phi.image(g)
    (1,2,3,4)
    sage: phi(g)
    (1,2,3,4)
    sage: phi.codomain()
    Dihedral group of order 8 as a permutation group
    sage: phi.codomain()
    Dihedral group of order 8 as a permutation group
    sage: phi.domain()
    Cyclic group of order 4 as a permutation group
"""

# ****************************************************************************
#       Copyright (C) 2006 David Joyner and William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.morphism import Morphism
from sage.groups.perm_gps.permgroup import PermutationGroup, PermutationGroup_generic


class PermutationGroupMorphism(Morphism):
    r"""
    A set-theoretic map between PermutationGroups.
    """
    def _repr_type(self):
        r"""
        Return the type of this morphism.

        This is used for printing the morphism.

        EXAMPLES::

            sage: G = PSL(2,7)
            sage: D, iota1, iota2, pr1, pr2 = G.direct_product(G)
            sage: pr1._repr_type()
            'Permutation group'
        """
        return "Permutation group"

    def kernel(self):
        r"""
        Return the kernel of this homomorphism as a permutation group.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: g = G([(1,2,3,4)])
            sage: phi = PermutationGroupMorphism_im_gens(G, H, [1])
            sage: phi.kernel()
            Subgroup generated by [(1,2,3,4)] of
             (Cyclic group of order 4 as a permutation group)

        ::

            sage: G = PSL(2,7)
            sage: D = G.direct_product(G)
            sage: H = D[0]
            sage: pr1 = D[3]
            sage: G.is_isomorphic(pr1.kernel())
            True
        """
        return self.domain().subgroup(gap_group=self._libgap_().Kernel())

    def image(self, J):
        r"""
        Compute the subgroup of the codomain `H` which is the image of `J`.

        `J` must be a subgroup of the domain `G`.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: g = G([(1,2,3,4)])
            sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens()))
            sage: phi.image(G)
            Subgroup generated by [(1,2,3,4)] of
             (Dihedral group of order 8 as a permutation group)
            sage: phi.image(g)
            (1,2,3,4)

        ::

            sage: G = PSL(2,7)
            sage: D = G.direct_product(G)
            sage: H = D[0]
            sage: pr1 = D[3]
            sage: pr1.image(G)
            Subgroup generated by [(3,7,5)(4,8,6), (1,2,6)(3,4,8)] of
             (The projective special linear group of degree 2 over Finite Field of size 7)
            sage: G.is_isomorphic(pr1.image(G))
            True

        Check that :issue:`28324` is fixed::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: f = x^4 + x^2 - 3
            sage: L.<a> = f.splitting_field()
            sage: G = L.galois_group()
            sage: D4 = DihedralGroup(4)
            sage: h = D4.isomorphism_to(G)
            sage: h.image(D4).is_isomorphic(G)
            True
            sage: all(h.image(g) in G for g in D4.gens())
            True
        """
        H = self.codomain()
        if J in self.domain():
            J = PermutationGroup([J])
            G = self._libgap_().Image(J)
            return H.subgroup(gap_group=G).gens()[0]
        else:
            G = self._libgap_().Image(J)
            return H.subgroup(gap_group=G)

    def __call__(self, g):
        r"""
        Some python code for wrapping GAP's ``Images`` function but only for
        permutation groups. This raises an error if g is not in G.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens()))
            sage: g = G([(1,3),(2,4)]); g
            (1,3)(2,4)
            sage: phi(g)
            (1,3)(2,4)
        """
        return self.image(g)


class PermutationGroupMorphism_id(PermutationGroupMorphism):
    pass


class PermutationGroupMorphism_from_gap(PermutationGroupMorphism):
    def __init__(self, G, H, gap_hom):
        r"""
        This is a Python trick to allow Sage programmers to create a group
        homomorphism using GAP using very general constructions. An example
        of its usage is in the direct_product instance method of the
        :class:`PermutationGroup_generic` class in permgroup.py.

        Basic syntax:

        ``PermutationGroupMorphism_from_gap(domain_group,
        range_group, 'phi:=gap_hom_command;', 'phi')``. And don't forget the
        line: ``from sage.groups.perm_gps.permgroup_morphism import
        PermutationGroupMorphism_from_gap`` in your program.

        EXAMPLES::

            sage: from sage.groups.perm_gps.permgroup_morphism import PermutationGroupMorphism_from_gap
            sage: G = PermutationGroup([[(1,2),(3,4)], [(1,2,3,4)]])
            sage: H = G.subgroup([G([(1,2,3,4)])])
            sage: PermutationGroupMorphism_from_gap(H, G, gap.Identity)
            Permutation group morphism:
              From: Subgroup generated by [(1,2,3,4)] of
                    (Permutation Group with generators [(1,2)(3,4), (1,2,3,4)])
              To:   Permutation Group with generators [(1,2)(3,4), (1,2,3,4)]
              Defn: Identity
        """
        if not all(isinstance(X, PermutationGroup_generic) for X in [G, H]):
            raise TypeError("the groups must be permutation groups")
        PermutationGroupMorphism.__init__(self, G, H)
        self._gap_hom = gap_hom

    def _repr_defn(self):
        r"""
        Return the definition of this morphism.

        This is used when printing the morphism.

        EXAMPLES::

            sage: from sage.groups.perm_gps.permgroup_morphism import PermutationGroupMorphism_from_gap
            sage: G = PermutationGroup([[(1,2),(3,4)], [(1,2,3,4)]])
            sage: H = G.subgroup([G([(1,2,3,4)])])
            sage: phi = PermutationGroupMorphism_from_gap(H, G, gap.Identity)
            sage: phi._repr_defn()
            'Identity'
        """
        return str(self._gap_hom).replace('\n', '')

    def _libgap_(self):
        r"""
        Return a GAP version of this morphism.

        EXAMPLES::

            sage: from sage.groups.perm_gps.permgroup_morphism import PermutationGroupMorphism_from_gap
            sage: G = PermutationGroup([[(1,2),(3,4)], [(1,2,3,4)]])
            sage: H = G.subgroup([G([(1,2,3,4)])])
            sage: phi = PermutationGroupMorphism_from_gap(H, G, gap.Identity)
            sage: phi._libgap_()
            Identity
        """
        return self._gap_hom

    def __call__(self, g):
        r"""
        Some python code for wrapping GAP's Images function but only for
        permutation groups. This returns an error if g is not in G.

        EXAMPLES::

            sage: G = PSL(2,7)
            sage: D = G.direct_product(G)
            sage: H = D[0]
            sage: pr1 = D[3]
            sage: [pr1(g) for g in G.gens()]
            [(3,7,5)(4,8,6), (1,2,6)(3,4,8)]
        """
        return self.codomain()(self._libgap_().Image(g))


class PermutationGroupMorphism_im_gens(PermutationGroupMorphism):
    def __init__(self, G, H, gens=None):
        r"""
        Some python code for wrapping GAP's ``GroupHomomorphismByImages``
        function but only for permutation groups. Can be expensive if G is
        large. This returns "fail" if gens does not generate ``self`` or if the map
        does not extend to a group homomorphism, ``self`` - ``other``.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens())); phi
            Permutation group morphism:
              From: Cyclic group of order 4 as a permutation group
              To:   Dihedral group of order 8 as a permutation group
              Defn: [(1,2,3,4)] -> [(1,2,3,4)]
            sage: g = G([(1,3),(2,4)]); g
            (1,3)(2,4)
            sage: phi(g)
            (1,3)(2,4)
            sage: images = ((4,3,2,1),)
            sage: phi = PermutationGroupMorphism_im_gens(G, G, images)
            sage: g = G([(1,2,3,4)]); g
            (1,2,3,4)
            sage: phi(g)
            (1,4,3,2)

        AUTHORS:

        - David Joyner (2006-02)
        """
        if not all(isinstance(X, PermutationGroup_generic) for X in [G, H]):
            raise TypeError("the groups must be permutation groups")
        PermutationGroupMorphism.__init__(self, G, H)
        self._images = [H(img) for img in gens]

    def _repr_defn(self):
        r"""
        Return the definition of this morphism.

        This is used when printing the morphism.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens()))
            sage: phi._repr_defn()
            '[(1,2,3,4)] -> [(1,2,3,4)]'
        """
        return "%s -> %s" % (list(self.domain().gens()), self._images)

    def _libgap_(self):
        r"""
        Return a GAP representation of this morphism.

        EXAMPLES::

            sage: G = CyclicPermutationGroup(4)
            sage: H = DihedralGroup(4)
            sage: phi = PermutationGroupMorphism_im_gens(G, H, map(H, G.gens()))
            sage: phi._libgap_()
            [ (1,2,3,4) ] -> [ (1,2,3,4) ]
        """
        return self.domain()._libgap_().GroupHomomorphismByImages(self.codomain(), self.domain().gens(), self._images)
